"""
Backends for the frontend API
"""

from __future__ import annotations
from typing import Optional, Any, Union, TypeVar, Generic, Type, Callable, Sequence, Dict, Tuple, List
import contextlib
import numpy
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim
from returnn.util.basic import BehaviorVersion
from .types import RawTensorTypes, ItemKeyType

T = TypeVar("T")  # tf.Tensor, torch.Tensor or so
T2 = TypeVar("T2")
S = TypeVar("S")  # any nested structure, can be None


class Backend(Generic[T]):
    """
    Abstract base class for the backend, operating on tensor type T, i.e. :class:`Tensor[T]`.

    This class and instances do not have any state,
    and all functions are staticmethod (or classmethod).
    """

    # class attribs set by derived classes
    RawTensorType: Type[T]
    is_tensorflow: bool = False  # whether this framework uses TensorFlow
    is_backend_raw_tensor_dim_tag_independent: bool = True  # whether raw tensors of backend are independent of Dim

    def __init__(self):
        raise Exception("do not instantiate this class")

    # --- functions to override

    @staticmethod
    def executing_eagerly() -> bool:
        """
        :return: whether we are in eager execution mode
        """
        raise NotImplementedError

    @staticmethod
    def get_tensor_dependencies(x: Tensor) -> Sequence[Tensor]:
        """
        :param x: tensor
        :return: list of all tensors which are inputs to `x`, ancestor tensors, dependencies.
            E.g. :func:`tf.Tensor.op.inputs`.
            This mostly makes sense for graph-based frameworks
            but eager-based frameworks might have this too with enabled gradient tape,
            as they should know the inputs.
        """
        raise NotImplementedError

    @staticmethod
    def get_tensor_consumers(x: Tensor) -> Sequence[Tensor]:
        """
        :param x: tensor
        :return: list of all tensors depending on `x`, descendant tensors, used by.
            E.g. :func:`tf.Tensor.consumers`.
            This mostly makes sense for graph-based frameworks
            but eager-based frameworks might have this too with enabled gradient tape,
            as they should know the consumers.
        """
        raise NotImplementedError

    @staticmethod
    def cond(pred: Tensor, true_fn: Callable, false_fn: Callable):
        """
        cond: conditional execution.

        Note that this does not need an implementation for eager-based frameworks
        (:func:`executing_eagerly` returns True),
        as the :func:`returnn.frontend.cond` function already covers that case.
        """
        # noinspection PyProtectedMember
        assert not pred._raw_backend.executing_eagerly(), "should not get here"
        raise NotImplementedError

    @staticmethod
    def while_loop(
        cond: Callable[[S], Union[bool, Tensor]],
        body: Callable[[S], S],
        initial: S,
    ) -> S:
        """while loop"""
        raise NotImplementedError

    @staticmethod
    def set_random_seed(seed: int):
        """
        :param seed:
        """
        raise NotImplementedError

    @staticmethod
    def get_random_state() -> Dict[str, bytes]:
        """
        :return: random state
        """
        raise NotImplementedError

    @staticmethod
    def set_random_state(state: Dict[str, bytes]):
        """
        :param state: as returned by :func:`get_random_state`.
            This might not always be successful (e.g. different hardware, different backend version),
            so the calling code should always have called set_random_seed before to have the random generators
            in a reasonable fallback state.
        """
        raise NotImplementedError

    @staticmethod
    def get_dtype_name_raw(raw_tensor: T) -> str:
        """
        :return: dtype of raw tensor, as string
        """
        raise NotImplementedError

    @staticmethod
    def as_dtype_raw(dtype_name: str) -> Any:
        """
        :param dtype_name: e.g. "float32"
        :return: dtype object
        """
        raise NotImplementedError

    @staticmethod
    def get_ndim_raw(raw_tensor: T) -> int:
        """
        :return: ndim of raw tensor. assumes it is known
        """
        raise NotImplementedError

    @staticmethod
    def get_shape_raw(raw_tensor: T) -> Union[T, Tuple[Union[int, T]]]:
        """
        :return: shape of raw tensor
        """
        raise NotImplementedError

    @staticmethod
    def get_shape_tuple_raw(raw_tensor: T) -> Tuple[Union[int, T]]:
        """
        :return: shape of raw tensor. assumes that ndim is known.
            In eager frameworks, all dims are int.
        """
        raise NotImplementedError

    @staticmethod
    def get_known_shape_raw(raw_tensor: T) -> Tuple[Optional[int]]:
        """
        :return: shape of raw tensor, int for static known, None otherwise. assumes that ndim is known.
            This will not create any ops.
            In eager frameworks, all dims are known.
        """
        raise NotImplementedError

    @staticmethod
    def set_known_shape_raw(raw_tensor: T, shape: Tuple[Optional[int]]) -> None:
        """
        Sets the known shape of the raw tensor.
        This is only supported in graph-based frameworks,
        and just performs a check in eager frameworks.
        """
        # Nothing for eager-based frameworks.

    @staticmethod
    def get_new_dim_raw(raw_tensor: T, axis: int, *, name: str) -> Dim:
        """
        :param raw_tensor:
        :param axis:
        :param name:
        :return: dim tag of axis
        """
        raise NotImplementedError

    @staticmethod
    def get_device(x: Tensor) -> Optional[str]:
        """
        :param x:
        :return: device, or none if unknown or logic not supported
        """
        # default implementation: ignore device
        return None

    @staticmethod
    def copy_to_device(x: Tensor, device: Optional[str]) -> Tensor:
        """
        :param x: tensor
        :param device: e.g. "cpu" or "gpu"
        :return: tensor on device
        """
        # default implementation: ignore device
        return x

    @staticmethod
    def fill_raw(shape: Union[Sequence[Union[int, T]], T], value: Union[Any, T]) -> T:
        """
        :param shape: shape
        :param value: scalar value to fill
        :return: raw tensor filled with value everywhere
        """
        raise NotImplementedError

    @staticmethod
    def compare_raw(a: T, kind: str, b: T) -> T:
        """
        :param a:
        :param kind: "equal", "less", "less_equal", "greater", "greater_equal", "not_equal"
        :param b:
        :return: a `kind` b
        """
        raise NotImplementedError

    @staticmethod
    def combine_raw(a: T, kind: str, b: T) -> T:
        """
        :param a:
        :param kind: "add", "sub", "mul", "truediv", "floordiv", "mod", "pow",
            "maximum", "minimum", "logical_and", "logical_or", "squared_difference"
        :param b:
        :return: a `kind` b
        """
        raise NotImplementedError

    @staticmethod
    def reshape_raw(raw_tensor: T, shape: Union[Sequence[Union[int, T]], T]) -> T:
        """
        :param raw_tensor: raw tensor
        :param shape: new shape
        :return: reshaped raw tensor
        """
        raise NotImplementedError

    @classmethod
    def squeeze_raw(cls, raw_tensor: T, axes: Sequence[int]) -> T:
        """
        :param raw_tensor: raw tensor
        :param axes: axes to squeeze
        :return: squeezed raw tensor
        """
        # Default implementation using reshape_raw.
        known_shape = cls.get_known_shape_raw(raw_tensor)
        assert all([known_shape[axis] == 1 for axis in axes])
        new_shape = [dim for a, dim in enumerate(cls.get_shape_tuple_raw(raw_tensor)) if a not in axes]
        return cls.reshape_raw(raw_tensor, new_shape)

    @staticmethod
    def transpose_raw(raw_tensor: T, perm: Sequence[int]) -> T:
        """
        :param raw_tensor: raw tensor
        :param perm: permutation
        :return: transposed raw tensor
        """
        raise NotImplementedError

    @staticmethod
    def make_output_tensor(tensor: Tensor, dims: Sequence[Dim], *, name: str) -> Tensor:
        """
        :param tensor:
        :param dims:
        :param name:
        :return: tensor with dims order like in dims
        """
        assert len(dims) == len(tensor.dims)
        tensor = tensor.copy_compatible_to_dims(dims)
        tensor = tensor.copy(name=name)
        return tensor

    @staticmethod
    def expand_dims_raw(raw_tensor: T, axis: int) -> T:
        """
        :param raw_tensor:
        :param axis:
        :return: raw tensor with new axis
        """
        raise NotImplementedError

    @staticmethod
    def expand_raw(raw_tensor: T, axis: int, dim: Union[int, T]) -> T:
        """
        :param raw_tensor:
        :param axis: shape[axis] must be 1
        :param dim: the new dim for shape[axis]
        :return: shape[axis] expands to dim.
            in PyTorch or other frameworks which support custom strides,
            this is an efficient view and not a copy.
        """
        raise NotImplementedError

    @staticmethod
    def copy(tensor: Tensor) -> Tensor:
        """copy"""
        raise NotImplementedError

    @staticmethod
    def cast_raw(raw_tensor: T, dtype: str) -> T:
        """
        :param raw_tensor:
        :param dtype: e.g. "float32"
        :return: raw tensor with dtype casted
        """
        raise NotImplementedError

    @staticmethod
    def cast(tensor: Tensor, dtype: str) -> Tensor:
        """
        :param tensor:
        :param dtype: e.g. "float32"
        :return: tensor with dtype casted
        """
        # Default implementation using cast_raw.
        res = tensor.copy_template()
        res.dtype = dtype
        if res.sparse_dim:
            if dtype.startswith("int") or dtype.startswith("uint"):
                pass
            elif dtype == "bool" and res.sparse_dim.dimension == 2:
                pass
            else:
                res.sparse_dim = None
        # noinspection PyProtectedMember
        res.raw_tensor = tensor._raw_backend.cast_raw(tensor.raw_tensor, dtype)
        return res

    @staticmethod
    def set_requires_gradient(tensor: Tensor):
        """
        :param tensor:
        """
        raise NotImplementedError

    @staticmethod
    def gradient(y: Tensor, x: Tensor) -> Tensor:
        """
        :param y:
        :param x:
        :return: gradient of y w.r.t. x
        """
        raise NotImplementedError

    @staticmethod
    def stop_gradient(tensor: Tensor) -> Tensor:
        """
        :param tensor:
        :return: tensor with stopped gradient
        """
        raise NotImplementedError

    @staticmethod
    def scaled_gradient(tensor: Tensor, scale: Union[float, Tensor]) -> Tensor:
        """
        :param tensor:
        :param scale:
        :return: tensor with scaled gradient
        """
        raise NotImplementedError

    @staticmethod
    def scaled_gradient_ext(
        x: Tensor,
        *,
        scale: Union[float, Tensor] = 1.0,
        shift: Optional[Union[float, Tensor]] = None,
        scale_shift_by_sum_over_axis: Optional[Dim] = None,
    ):
        """
        :param x:
        :param scale: will scale gradient by this value
        :param shift: will shift gradient by this value
        :param scale_shift_by_sum_over_axis: if given, will scale and shift by the sum over the given axis
        :return: just x, but gradient in backward pass will be transformed accordingly
        """
        raise NotImplementedError

    @staticmethod
    def merge_dims(
        source: Tensor,
        *,
        dims: Sequence[Dim],
        out_dim: Optional[Dim] = None,
    ) -> Tuple[Tensor, Dim]:
        """
        Merges a list of axes into a single one. (Flatten the dims.)
        E.g. input is (batch, width, height, dim) and dims=(width,height), then we get (batch, width*height, dim).
        Or input is (batch, time, height, dim) and axes=(height,dim), then we get (batch, time, height*dim).

        :param source:
        :param dims:
        :param out_dim:
        :return: tensor, out_dim
        """
        raise NotImplementedError

    @staticmethod
    def split_dims(
        source: Tensor,
        *,
        axis: Dim,
        dims: Sequence[Dim],
        pad_to_multiples: Optional[bool] = None,
        pad_value: Union[None, int, float] = None,
    ) -> Tensor:
        """
        :param source:
        :param axis:
        :param dims:
        :param pad_to_multiples:
        :param pad_value:
        :return: source with axis replaced by dims
        """
        raise NotImplementedError

    @staticmethod
    def reshape(source: Tensor, in_dims: Sequence[Dim], out_dims: Sequence[Dim]) -> Tensor:
        """
        :param source: e.g. (..., old_dims, ...)
        :param in_dims: the old dims which should be reshaped into new_dims.
            This should only cover those dims which should be reshaped,
            not all the dims of the source.
        :param out_dims: the new dims which should be reshaped from old_dims.
            This is excluding any of the other dims in the source.
        :return: e.g. (..., new_dims, ...)
        """
        raise NotImplementedError

    @staticmethod
    def split(source: Tensor, *, axis: Dim, out_dims: Sequence[Dim]) -> Tuple[Tensor, ...]:
        """
        Split the input on the specified axis (by default feature).
        Basically a wrapper around tf.split.

        :param source: {..., axis}
        :param axis: some static axis
        :param out_dims: list of dims where sum(out_dims) == axis
        :return: tuple of tensors, same amount as out_dims,
            with the same shape as source, but with the specified axis replaced by the out_dims
        """
        raise NotImplementedError

    @staticmethod
    def expand_dim(source: Tensor, dim: Dim) -> Tensor:
        """
        :param source:
        :param dim:
        :return: source with dim added
        """
        raise NotImplementedError

    @staticmethod
    def squeeze(source: Tensor, axis: Dim) -> Tensor:
        """
        :param source:
        :param axis:
        :return: source with axis removed
        """
        raise NotImplementedError

    @staticmethod
    def concat(
        *sources: Tuple[Tensor, Dim],
        allow_broadcast: bool = False,
        out_dim: Dim,
    ) -> Tensor:
        """concat"""
        raise NotImplementedError

    @staticmethod
    def pad(
        source: Tensor,
        *,
        axes: Sequence[Dim],
        padding: Sequence[Tuple[Union[Dim, int], Union[Dim, int]]],
        out_dims: Sequence[Dim],
        mode: str = "constant",
        value: Optional[Union[rf.RawTensorTypes, Tensor]] = None,
    ) -> Tensor:
        """
        :param source:
        :param axes:
        :param padding:
        :param out_dims:
        :param mode:
        :param value:
        :return: padded tensor
        """
        raise NotImplementedError

    @staticmethod
    def cum_concat_step(source: Tensor, *, prev_accum: Tensor, axis: Dim, out_spatial_dim: Dim) -> Tensor:
        """
        Concatenates all previous frames over a time-axis.
        See RETURNN :class:`CumConcatLayer` for details.

        :param source: same dims as prev_accum except for the accum axis
        :param prev_accum: previous accumulated tensor, shape {..., axis}
        :param axis: the axis to accumulate over
        :param out_spatial_dim: the spatial dim of the output will be this dim. like axis+1.
        :return: accumulated. accumulated shape {..., out_spatial_dim},
            same shape as prev_accum with axis replaced by out_spatial_dim.
        """
        raise NotImplementedError

    # Restrict the possible activation function names,
    # to not get unexpected behavior,
    # or unwanted incompatibilities.
    _AllowedActivationFuncs = {
        "exp",
        "expm1",
        "log",
        "log1p",
        "sqrt",
        "rsqrt",
        "square",
        "abs",
        "tanh",
        "sigmoid",
        "log_sigmoid",
        "sin",
        "cos",
        "ceil",
        "floor",
        "round",
        "relu",
        "elu",
        "selu",
        "silu",
        "logical_not",
        "neg",
        "reciprocal",
    }

    @staticmethod
    def activation(tensor: Tensor, func: str) -> Tensor:
        """
        :param tensor:
        :param func: "tanh", "sigmoid", "relu", ...
        :return: tensor with elementwise activation applied
        """
        out = tensor.copy_template(name=func)
        # noinspection PyProtectedMember
        out_raw = tensor._raw_backend.activation_raw(tensor.raw_tensor, func)
        # noinspection PyProtectedMember
        out.dtype = tensor._raw_backend.get_dtype_name_raw(out_raw)
        out.raw_tensor = out_raw
        return out

    @staticmethod
    def activation_raw(raw_tensor: T, func: str) -> T:
        """
        :param raw_tensor:
        :param func: "tanh", "sigmoid", "relu", ...
        :return: raw tensor with elementwise activation applied
        """
        raise NotImplementedError

    @staticmethod
    def safe_log(tensor: Tensor, *, eps: float) -> Tensor:
        """
        :param tensor:
        :param eps:
        :return: log(tensor + eps) in the default case. but some backends might do more things,
            like if tensor = softmax(logits), then this would be log_softmax(logits) instead.
        """
        return rf.log(rf.maximum(tensor, eps))

    @staticmethod
    def softmax(tensor: Tensor, *, axis: Dim, use_mask: bool = True) -> Tensor:
        """
        :param tensor:
        :param axis:
        :param use_mask:
        :return: softmax over axis
        """
        raise NotImplementedError

    @staticmethod
    def log_softmax(tensor: Tensor, *, axis: Dim, use_mask: bool = True) -> Tensor:
        """
        :param tensor:
        :param axis:
        :param use_mask:
        :return: log_softmax over axis
        """
        raise NotImplementedError

    @staticmethod
    def softmax_cross_entropy_with_logits(*, logits: Tensor, targets: Tensor, axis: Dim):
        """
        Efficient cross entropy.

        :param logits: target estimates given as inputs to softmax (i.e. unnormalized)
        :param targets: probabilities, i.e. normalized, can also be sparse
        :param axis: class labels dim over which softmax is computed
        :return: cross entropy (same Dims as 'logits' but without 'axis')
        """
        raise NotImplementedError

    @staticmethod
    def ctc_loss(
        *,
        logits: Tensor,
        targets: Tensor,
        input_spatial_dim: Dim,
        targets_spatial_dim: Dim,
        blank_index: int,
        max_approx: bool = False,
    ) -> Tensor:
        """
        Calculates the CTC loss.
        """
        raise NotImplementedError

    @staticmethod
    def have_sequence_mask_raw() -> bool:
        """
        :return: whether we have a sequence_mask_raw implementation
        """
        return False

    @staticmethod
    def sequence_mask_raw(lengths: T, *, batch_major: bool = True) -> T:
        """
        Like tf.sequence_mask().

        :param lengths: shape (batch,)
        :param batch_major:
        :return: tensor mask of shape (batch,maxlen) if batch_major else (maxlen,batch) of type bool
        """
        raise NotImplementedError

    @staticmethod
    @contextlib.contextmanager
    def name_scope_raw(name: str) -> Any:
        """
        Default implementation for eager-based frameworks:
        Do nothing, tensors do not have a name.

        :param name:
        :return: context manager
        """
        # Default implementation for eager-based frameworks
        yield  # nothing to do

    @staticmethod
    @contextlib.contextmanager
    def control_dependencies_raw(dependencies: Sequence[Any]) -> Any:
        """
        Default implementation for eager-based frameworks:
        Do nothing, we expect that the dependencies are already executed.

        :param dependencies: raw tensors or ops
        :return: context manager
        """
        # Default implementation for eager-based frameworks
        yield

    @staticmethod
    def identity_with_control_dependencies_raw(raw_tensor: T, dependencies: Sequence[Any]) -> T:
        """
        Default implementation for eager-based frameworks:
        Do nothing, we expect that the dependencies are already executed.

        :param raw_tensor: raw tensor
        :param dependencies: raw tensors or ops
        :return: raw tensor
        """
        # Default implementation for eager-based frameworks
        return raw_tensor

    @staticmethod
    def create_placeholder_raw(tensor: Tensor) -> T:
        """
        :return: tf.placeholder in TF

        This is really only for TensorFlow for the deprecated option auto_create_placeholders
        and should not be used in other backends,
        even in graph-based backends.
        Rather, the logic to create placeholders should be done elsewhere.
        """
        raise Exception("create_placeholder not supported by backend")

    @staticmethod
    def create_parameter_raw(tensor: rf.Parameter, *, device: Optional[str] = None) -> T:
        """
        :return: parameter (by default trainable)
        """
        raise NotImplementedError

    @staticmethod
    def set_parameter_initial_value(param: rf.Parameter, value: Union[None, Tensor, rf.RawTensorTypes]) -> None:
        """
        :param param: parameter
        :param value: initial value
        """
        raise NotImplementedError

    @staticmethod
    def set_parameter_trainable(param: rf.Parameter, trainable: bool) -> None:
        """
        :param param: parameter
        :param trainable: whether the parameter should be trainable
        """
        raise NotImplementedError

    @staticmethod
    def parameter_assign(param: rf.Parameter, value: Tensor, *, op: str = "assign") -> None:
        """
        :param param: parameter
        :param value: new value
        :param op: "assign" or "add"
        """
        raise NotImplementedError

    @staticmethod
    def parameter_assign_key(
        param: rf.Parameter,
        key: ItemKeyType,
        value: Tensor,
        *,
        op: str = "assign",
        axis: Optional[Union[Dim, Sequence[Dim]]] = None,
        key_dim: Union[None, Dim, Sequence[Union[None, Dim]]] = None,
    ) -> None:
        """
        :param param: parameter
        :param key: optional key for slice assign, like var[key] = value or var[key] += value.
        :param value: new value
        :param op: "assign" or "add"
        :param axis: if key is given, this axis is used.
            if key are indices (without specified sparse_dim), axis must be specified.
        :param key_dim: resulting dim after slicing with key
        """
        raise NotImplementedError

    @staticmethod
    def parameter_move_to(param: rf.Parameter, *, device: Optional[str] = None, dtype: Optional[str] = None):
        """
        Updates `param` inplace, but `param.raw_tensor` might be a new instance.

        :param param:
        :param device:
        :param dtype:
        """
        raise NotImplementedError

    @staticmethod
    def runtime_sanity_checks(tensor: Tensor) -> Any:
        """
        Checks whether the tensor.raw_tensor is consistent with the tensor metadata.

        In graph-based frameworks (TF graph), we return some operation here.
        In eager frameworks, we would not return anything but instead directly perform the checks.
        """
        # By default, we do not do any checks. This is optional for the backend.
        pass

    @staticmethod
    def is_valid_in_current_graph(tensor: Tensor) -> bool:
        """
        :return: whether the raw tensor is valid in the current graph.
            In eager-mode frameworks, this is always true -- there is no graph.
        """
        return True

    @staticmethod
    def format_graph_output(raw_tensor: T, *, max_depth: Optional[int] = None) -> str:
        """
        :return: the computation graph leading to this tensor formatted.
            In eager-mode frameworks, this is not supported and returns None.
        """
        return "<no-graph>"

    @staticmethod
    def convert_to_tensor(
        value: Union[Tensor, T, RawTensorTypes],
        *,
        dims: Sequence[Dim],
        dtype: str,
        sparse_dim: Optional[Dim] = None,
        device: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Tensor[T]:
        """
        :param value: tensor, or scalar raw tensor or some other scalar value
        :param dims:
        :param dtype:
        :param sparse_dim:
        :param device:
        :param name:
        :return: tensor
        """
        raise NotImplementedError

    @staticmethod
    def full(
        dims: Sequence[Dim],
        fill_value: Union[RawTensorTypes, Tensor],
        *,
        dtype: str,
        device: Optional[str] = None,
        sparse_dim: Optional[Dim] = None,
        feature_dim: Optional[Dim] = None,
    ) -> Tensor:
        """
        https://data-apis.org/array-api/latest/API_specification/generated/array_api.full.html

        :param dims:
        :param fill_value:
        :param dtype:
        :param device:
        :param sparse_dim:
        :param feature_dim:
        :return: tensor
        """
        raise NotImplementedError

    @classmethod
    def compare(
        cls,
        a: Union[Tensor, RawTensorTypes],
        kind: str,
        b: Union[Tensor, RawTensorTypes],
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """compare, default implementation using compare_raw"""
        from . import _utils

        out, a_raw, b_raw = _utils.bin_op_out_template(
            cls,
            a,
            b,
            name=kind,
            copy_sparse_dim=False,
            allow_broadcast_all_sources=allow_broadcast_all_sources,
            dim_order=dim_order,
        )
        out_raw = cls.compare_raw(a_raw, kind, b_raw)
        out.dtype = cls.get_dtype_name_raw(out_raw)
        out.raw_tensor = out_raw
        return out

    @classmethod
    def combine(
        cls,
        a: Union[Tensor, RawTensorTypes],
        kind: str,
        b: Union[Tensor, RawTensorTypes],
        *,
        allow_broadcast_all_sources: Optional[bool] = None,
        dim_order: Optional[Sequence[Dim]] = None,
    ) -> Tensor:
        """combine, default implementation using combine_raw"""
        from . import _utils

        out, a_raw, b_raw = _utils.bin_op_out_template(
            cls,
            a,
            b,
            name=kind,
            allow_broadcast_all_sources=allow_broadcast_all_sources,
            dim_order=dim_order,
        )
        out_raw = cls.combine_raw(a_raw, kind, b_raw)
        out.dtype = cls.get_dtype_name_raw(out_raw)
        out.raw_tensor = out_raw
        return out

    @staticmethod
    def gather(
        source: Tensor,
        *,
        indices: Union[Tensor, int],
        axis: Dim,
        clip_to_valid: bool = False,
    ) -> Tensor:
        """
        Gathers slices on a specified axis from the source using indices.
        If the source is of the shape ``[B,D,F1]``, and indices of shape ``[B,F2]``,
        this will yield output of the shape ``[B,F2,F1]`` where

        ``output[b,f2,f1] = source[b,indices[b,f2],f1]``

        (if ``D`` is the axis to gather from).
        In general, all shared axes of the input and the positions will be considered as batch-axes.

        The ``indices`` argument can also be an ``int``.
        In this case, this simply gives ``source[indices]`` on the specified ``axis``.

        :param source:
        :param indices: indices used to select the slices of the source from.
            If another tensor, must be of type ``int32`` or ``int64``.
            Can also specify a constant ``int``.
        :param axis: The axis into which we gather the indices into
        :param clip_to_valid: if True, the indices will be clipped to the valid range of the input
            Also taking seq lengths into account.
        :return: gathered values
        """
        raise NotImplementedError

    @staticmethod
    def scatter(
        source: Tensor,
        *,
        indices: Tensor,
        indices_dim: Union[Dim, Sequence[Dim]],
        out_dim: Union[Dim, Sequence[Dim]],
    ) -> Tensor:
        """
        Scatters into new zero-tensor.
        If entries in indices are duplicated, the corresponding values in source will be added together
        (scatter_add in PyTorch).
        (TF segment_sum can be implemented via this.)

        :param source: [batch_dims..., indices_dim(s)..., feature_dims...]
        :param indices: [batch_dims..., indices_dim(s)...] -> out_dim
        :param indices_dim:
        :param out_dim:
        :return: [batch_dims..., out_dim, feature_dims...]
        """
        raise NotImplementedError

    @staticmethod
    def slice(
        source: Tensor,
        *,
        axis: Dim,
        start: Optional[Union[int, Tensor]] = None,
        end: Optional[Union[int, Tensor]] = None,
        step: Optional[Union[int, Tensor]] = None,
        size: Optional[Union[int, Tensor, Dim]] = None,
        out_dim: Dim,
    ) -> Tensor:
        """slice"""
        raise NotImplementedError

    @staticmethod
    def where(
        cond: Tensor,
        true_: Union[Tensor, rf.RawTensorTypes],
        false_: Union[Tensor, rf.RawTensorTypes],
        *,
        allow_broadcast_all_sources: bool = False,
    ) -> Tensor:
        """where"""
        raise NotImplementedError

    @staticmethod
    def clip_by_value(
        x: Tensor,
        clip_value_min: Union[Tensor, rf.RawTensorTypes],
        clip_value_max: Union[Tensor, rf.RawTensorTypes],
        *,
        allow_broadcast_all_sources: bool = False,
    ) -> Tensor:
        """clip by value"""
        raise NotImplementedError

    @staticmethod
    def matmul(a: Tensor[T], b: Tensor[T], *, reduce: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
        """
        This performs a batched matmul of two sources a and b
        (non-batched matmul and dot product are special cases).
        The underlying operation is a batched matmul (shared..., I, J) * (shared..., J, K) -> (shared..., I, K).
        The inputs a and b are transformed internally into the required shapes in the following way:
        The axis J is specified via the Dim given as 'reduce'. If multiple reduce Dims are given the corresponding axes
        are merged into one before the matmul via a reshape. All other matching Dims in a and b will be treated as
        batch dimensions ('shared...'). Dims unique to a and b define the axes I and K, respectively. (Multiple or no
        unique axes in a and b are supported too.)

        Depending on which Dims exist in a, b and reduce this dot operation can be used to compute scaling, scalar
        product, outer product, matrix-vector multiplication, matrix-matrix multiplication etc. (all possibly batched).

        :param a:
        :param b:
        :param reduce: Dims over which to perform the product, have to be present in both a and b
        :param use_mask: If the reduction is over dynamic axes, to get the correct sum reduction,
            we need to apply masking to one of the inputs. This is done automatically.
            By disabling this flag, this would be disabled.
        :return: result of dot product, Dim order: common axes as sorted in a, unique axes of a (in order),
            unique axes of b (in order)
        """
        raise NotImplementedError

    @staticmethod
    def range_over_dim(dim: Dim, *, dtype: Optional[str] = None, device: Optional[str] = None) -> Tensor[T]:
        """
        :param dim:
        :param dtype:
        :param device:
        :return: tensor with shape [dim]
        """
        raise NotImplementedError

    @staticmethod
    def replace_dim(source: Tensor, *, in_dim: Dim, out_dim: Dim) -> Tensor:
        """
        :param source:
        :param in_dim:
        :param out_dim:
        :return: source with in_dim replaced by out_dim.
        """
        # This default implementation works fine as long as the backend
        # does not have special treatments of Tensor and dim tags itself (like TF net dict backend).
        if not out_dim.is_dim_known():
            out_dim.copy_from(in_dim)
        out = source.copy_template_replace_dim_tag(
            axis=source.get_axis_from_description(in_dim), new_dim_tag=out_dim, name="replace_dim"
        )
        out.raw_tensor = source.raw_tensor
        return out

    _AllowedReduceModes = {"sum", "max", "min", "mean", "logsumexp", "any", "all", "argmin", "argmax"}

    @staticmethod
    def reduce(
        source: Tensor[T],
        *,
        mode: str,
        axis: Union[Dim, Sequence[Dim]],
        use_mask: bool = True,
    ) -> Tensor[T]:
        """
        Reduce the tensor along the given axis

        :param source:
        :param mode: "sum", "max", "min", "mean", "logsumexp", "any", "all", "argmin", "argmax"
        :param axis:
        :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
        :return: tensor with axis removed
        """
        raise NotImplementedError

    # noinspection PyShadowingBuiltins
    @staticmethod
    def top_k(
        source: Tensor,
        *,
        axis: Union[Dim, Sequence[Dim]],
        k: Union[int, Tensor],
        k_dim: Optional[Dim] = None,
        sorted: bool = True,
    ) -> Tuple[Tensor, Union[Tensor, Sequence[Tensor]], Dim]:
        """top_k. see :func:`top_k`"""
        raise NotImplementedError

    @staticmethod
    def random(
        *,
        dims: Sequence[Dim],
        dtype: str,
        device: Optional[str] = None,
        sparse_dim: Optional[Dim] = None,
        feature_dim: Optional[Dim] = None,
        distribution: str,
        mean: Optional[Union[int, float, Tensor]] = None,
        stddev: Optional[Union[int, float, Tensor]] = None,
        bound: Optional[Union[int, float, Tensor]] = None,
        minval: Optional[Union[int, float, Tensor]] = None,
        maxval: Optional[Union[int, float, Tensor]] = None,
        seed: Optional[Union[int, Sequence[int], numpy.ndarray]] = None,
        algorithm: Optional[str] = None,
        explicit_state: Optional[Tensor] = None,
        auto_update_state: Optional[bool] = None,
        static: Optional[bool] = None,
        out: Optional[Tensor] = None,
    ) -> Tensor:
        """
        random. See `rf.random` for details.
        """
        raise NotImplementedError

    @staticmethod
    def masked_select(
        tensor: Tensor, *, mask: Tensor, dims: Sequence[Dim], out_dim: Optional[Dim] = None
    ) -> Tuple[Tensor, Dim]:
        """
        :param tensor:
        :param mask:
        :param dims: the order of the dims defines the format. those dims should be exactly the dims of the mask.
        :param out_dim:
        :return: tensor where all dims in mask/dims are removed and replaced by a new dim.
            the new dim is also returned.
            if mask==True for all elements, the returned tensor would be simply the flattened input tensor.
        """
        raise NotImplementedError

    @staticmethod
    def masked_scatter(source: Tensor, *, mask: Tensor, dims: Sequence[Dim], in_dim: Dim) -> Tensor:
        """
        The inverse of :func:`masked_select`.

        :param source: [in_dim, F...]
        :param mask: [dims...] -> bool (e.g. [B,T])
        :param dims: the order of the dims defines the format. those dims should be exactly the dims of the mask.
        :param in_dim: the dim of the source which should be scattered into the mask.
        :return: [dims..., F...]
        """
        raise NotImplementedError

    @staticmethod
    def batch_norm(
        source: Tensor,
        *,
        in_dim: Union[Dim, Sequence[Dim]],
        running_mean: Tensor,
        running_variance: Tensor,
        gamma: Optional[Tensor],
        beta: Optional[Tensor],
        epsilon: float,
        momentum: float,
        affine: bool,
        use_mask: bool,
    ) -> Tensor:
        """
        :param source:
        :param in_dim:
        :param running_mean:
        :param running_variance:
        :param gamma:
        :param beta:
        :param epsilon:
        :param momentum:
        :param affine:
        :param use_mask:
        :return:
        """
        raise NotImplementedError

    # noinspection PyShadowingBuiltins
    @staticmethod
    def conv(
        source: Tensor,
        *,
        in_dim: Dim,
        out_dim: Dim,
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
        filter: Tensor,
        filter_size: Sequence[Dim],  # to have the order well-defined
        padding: str,
        strides: Optional[Union[int, Sequence[int]]] = None,
        dilation_rate: Optional[Union[int, Sequence[int]]] = None,
        groups: Optional[int] = None,
        bias: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        """convolution"""
        raise NotImplementedError

    # noinspection PyShadowingBuiltins
    @staticmethod
    def transposed_conv(
        source: Tensor,
        *,
        in_dim: Dim,
        out_dim: Dim,
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
        filter: Tensor,
        filter_size: Sequence[Dim],
        padding: str,
        remove_padding: Union[Sequence[int], int] = 0,
        output_padding: Optional[Union[Sequence[Optional[int]], int]] = None,
        strides: Optional[Sequence[int]] = None,
        bias: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        """transposed convolution"""
        raise NotImplementedError

    @staticmethod
    def pool(
        source: Tensor,
        *,
        mode: str,
        pool_size: Sequence[int],
        padding: str = "valid",
        dilation_rate: Union[Sequence[int], int] = 1,
        strides: Sequence[int],
        in_spatial_dims: Sequence[Dim],
        out_spatial_dims: Optional[Sequence[Dim]] = None,
    ) -> Tuple[Tensor, Sequence[Dim]]:
        """pooling"""
        raise NotImplementedError

    @staticmethod
    def stft(
        x: Tensor,
        *,
        in_spatial_dim: Dim,
        frame_step: int,
        frame_length: int,
        fft_length: int,
        window_use_frame_length: bool = True,
        align_window_left: bool = True,
        window_enforce_even: bool = True,
        out_spatial_dim: Dim,
        out_dim: Dim,
    ) -> Tensor:
        """stft. see :func:`stft` for details."""
        raise NotImplementedError

    @staticmethod
    def lstm(
        source: Tensor,
        *,
        state_h: Tensor,
        state_c: Tensor,
        ff_weight: Tensor,
        rec_weight: Tensor,
        bias: Tensor,
        spatial_dim: Dim,
        in_dim: Dim,
        out_dim: Dim,
    ) -> Tuple[Tensor, Tuple[Tensor, Tensor]]:
        """
        Functional LSTM.

        :param source: Tensor of shape [*, in_dim].
        :param state_c:
        :param state_h:
        :param ff_weight: Parameters for the weights of the feed-forward part.
        :param rec_weight: Parameters for the weights of the recurrent part.
        :param bias: Parameters for the bias.
        :param spatial_dim: Dimension in which the LSTM operates.
        :param in_dim:
        :param out_dim:
        :return: output, (state_h, state_c)
        """
        raise NotImplementedError

    # For eager-based backends, this is a reasonable default implementation and type.
    TensorArrayType = List[Tensor]

    @classmethod
    def tensor_array_create(cls) -> TensorArrayType:
        """
        :return: empty TensorArray
        """
        if cls.executing_eagerly():
            return []
        raise NotImplementedError

    @staticmethod
    def tensor_array_unstack(tensor: Tensor, *, axis: Dim) -> TensorArrayType:
        """
        :param tensor:
        :param axis:
        :return: list of tensors
        """
        raise NotImplementedError

    @staticmethod
    def tensor_array_stack(tensor_array: TensorArrayType, *, axis: Dim, tensor_template: Tensor) -> Tensor:
        """
        :param tensor_array:
        :param axis:
        :param tensor_template: per element shape, excluding axis
        :return: tensor
        """
        raise NotImplementedError

    @classmethod
    def tensor_array_push_back(cls, tensor_array: TensorArrayType, value: Tensor) -> TensorArrayType:
        """
        :param tensor_array:
        :param value:
        :return: tensor_array
        """
        if cls.executing_eagerly():
            tensor_array.append(value)
            return tensor_array
        raise NotImplementedError

    @classmethod
    def tensor_array_get_item(cls, tensor_array: TensorArrayType, index: Union[int, Tensor]) -> Tensor:
        """
        :param tensor_array:
        :param index:
        :return: tensor
        """
        if cls.executing_eagerly():
            if isinstance(index, Tensor):
                assert index.dims == (), f"index {index} must be scalar"
                index = int(index.raw_tensor)
            return tensor_array[index]
        raise NotImplementedError


# We use a global instance, and we modify __class__ inplace,
# such that any reference to this can be updated.
# This is exposed to the user as `returnn.frontend`.
# The __class__ assignment is done in `select_engine`.
# Use object.__new__ because we disallow creating instances of Frontend.
global_backend = object.__new__(Backend)

_backend_tensor_type_dispatch_table = {}  # type: Dict[Type, Type[Backend]]


def select_backend_tf():
    """
    Selects the RETURNN layers backend (based on TF).
    """
    import tensorflow as tf

    backend = get_backend_by_raw_tensor_type(tf.Tensor)  # side-effect: register it
    global_backend.__class__ = backend
    BehaviorVersion.set_min_behavior_version(16)


def select_backend_returnn_layers_tf():
    """
    Selects the RETURNN layers backend (based on TF).
    """
    from returnn.tf.frontend_layers import Layer

    backend = get_backend_by_raw_tensor_type(Layer)  # side-effect: register it
    global_backend.__class__ = backend


def select_backend_torch():
    """
    Selects the PyTorch (low-level) backend.
    """
    import torch

    backend = get_backend_by_raw_tensor_type(torch.Tensor)  # side-effect: register it
    global_backend.__class__ = backend
    BehaviorVersion.set_min_behavior_version(16)

    from returnn.frontend import _native

    _native.setup()
    _native.setup_torch()


def get_backend_by_tensor(tensor: Tensor, *, fallback: Optional[T2] = None) -> Union[Type[Backend[T]], T2]:
    """
    :param tensor:
    :param fallback:
    """
    if fallback and tensor.raw_tensor is None:
        return fallback
    assert tensor.raw_tensor is not None
    return get_backend_by_raw_tensor_type(type(tensor.raw_tensor))


def get_backend_by_raw_tensor_type(tensor_type: Type[T]) -> Union[Type[Backend[T]]]:
    """
    :param tensor_type:
    """
    if tensor_type in _backend_tensor_type_dispatch_table:  # fast path
        return _backend_tensor_type_dispatch_table[tensor_type]

    if not isinstance(tensor_type, type):
        raise TypeError(f"Expected type, got {tensor_type!r} of type {type(tensor_type)}")
    tensor_type: Type[T]

    # We don't register all possible subclasses in the dispatch table.
    # Check through the MRO.
    for base_type in tensor_type.__mro__:
        if base_type in _backend_tensor_type_dispatch_table:
            # Also register it for faster future lookups.
            register_backend_by_tensor_type(tensor_type, _backend_tensor_type_dispatch_table[base_type])
            return _backend_tensor_type_dispatch_table[base_type]

    # It would be registered if there was any select_engine or select_backend_* call.
    # However, some code might not have done that, so for the common cases,
    # we do it here.
    for base_type in tensor_type.__mro__:
        if base_type.__module__.split(".")[0] == "tensorflow":
            from returnn.tf.frontend_low_level import TFBackend

            backend_type = TFBackend
            tensor_types = _get_tensor_types_tf()
        elif base_type.__module__.split(".")[0] == "torch":
            from returnn.torch.frontend import TorchBackend

            backend_type = TorchBackend
            tensor_types = _get_tensor_types_torch()
        elif base_type.__module__.startswith("returnn.tf.frontend_layers."):
            from returnn.tf.frontend_layers import ReturnnLayersBackend, Layer

            backend_type = ReturnnLayersBackend
            tensor_types = (Layer,)
        elif issubclass(base_type, numpy.ndarray):
            from ._numpy_backend import NumpyBackend

            backend_type = NumpyBackend
            tensor_types = (numpy.ndarray,)
        else:
            continue

        assert any(
            issubclass(base_type, type_) for type_ in tensor_types
        ), f"tensor type {tensor_type} base_type {base_type} not in {tensor_types}, expected for backend {backend_type}"
        for base_type_ in tensor_types:
            register_backend_by_tensor_type(base_type_, backend_type)
        return backend_type

    raise TypeError(f"unknown tensor type {tensor_type} with mro {tensor_type.__mro__}")


def register_backend_by_tensor_type(tensor_type: Type[T], backend: Type[Backend[T]]):
    """
    :param tensor_type:
    :param backend:
    """
    _backend_tensor_type_dispatch_table[tensor_type] = backend


def _get_tensor_types_tf():
    """
    :return: tuple of relevant tensor types in TF.
        Note that it is not so important to cover all, as we also check issubclass as a fallback.
    """
    import tensorflow as tf

    ls = [tf.Tensor, tf.Variable]
    return tuple(ls)


def _get_tensor_types_torch():
    """
    :return: tuple of relevant tensor types in PyTorch.
        Note that it is not so important to cover all, as we also check issubclass as a fallback.
    """
    import torch

    ls = [torch.Tensor, torch.nn.Parameter]
    return tuple(ls)



"""
tests if the graph values of two different training runs on are equal
can be used to test determinism of demo training on GPU and CPU.
in some cases the test will give a false negative (on GPU)

state 16.04.19 following demos are confirmed not deterministic on GPU:
with: CUDA 10.0, tf: 1.13.0 nvcc:

- demos/demo-tf-vanilla-lstm.12ax.config
- demos/demo-tf-native-lstm-lowmem.12ax.config

- demos/demo-tf-native-lstm.12ax.config

this test can be used on all lstm demos:

to run this test you might need to apply workaround: https://github.com/rwth-i6/returnn/issues/87
Example:

  python3 tests/test_TF_determinismus.py "<democonfig-location>"



"""

import sys
import os

import _setup_test_env  # noqa
from returnn.tf.engine import *
from returnn.config import Config

from numpy.testing import assert_array_equal, assert_equal

config = Config()
if len(sys.argv) == 1:
    config.load_file("demos/demo-tf-vanilla-lstm.12ax.config")
else:
    print(str(sys.argv[1]))
    config.load_file(sys.argv[1])


def load_data():
    from returnn.__main__ import load_data

    dev_data, _ = load_data(
        config, 0, "dev", chunking=config.value("chunking", ""), seq_ordering="sorted", shuffle_frames_of_nseqs=0
    )
    eval_data, _ = load_data(
        config, 0, "eval", chunking=config.value("chunking", ""), seq_ordering="sorted", shuffle_frames_of_nseqs=0
    )
    train_data, _ = load_data(config, 0, "train")
    return dev_data, eval_data, train_data


def test_determinism_of_vanillalstm():
    def create_engine():
        dev_data, eval_data, train_data = load_data()

        engine = Engine()
        engine.init_train_from_config(config, train_data, dev_data, eval_data)
        engine.init_train_epoch()
        engine.train_batches = engine.train_data.generate_batches(
            recurrent_net=engine.network.recurrent,
            batch_size=engine.batch_size,
            max_seqs=engine.max_seqs,
            max_seq_length=engine.max_seq_length,
            seq_drop=engine.seq_drop,
            shuffle_batches=engine.shuffle_batches,
            used_data_keys=engine.network.used_data_keys,
        )
        engine.updater.set_learning_rate(engine.learning_rate, session=engine.tf_session)
        engine.updater.init_optimizer_vars(session=engine.tf_session)
        return engine

    def train_engine_fetch_vars(engine):
        data_provider = engine._get_data_provider(
            dataset=engine.train_data, batches=engine.train_batches, feed_dict=True
        )
        feed_dict, _ = data_provider.get_feed_dict(single_threaded=True)
        trainer = Runner(engine=engine, dataset=engine.train_data, batches=engine.train_batches, train=True)
        feed_dict, _ = data_provider.get_feed_dict(single_threaded=True)
        trainer.run(report_prefix="One Run")

        return [e.eval(engine.tf_session) for e in engine.network.get_params_list()]

    e1 = create_engine()
    r1 = train_engine_fetch_vars(e1)

    e2 = create_engine()
    r2 = train_engine_fetch_vars(e2)

    assert_array_equal(r1, r2)


test_determinism_of_vanillalstm()



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Various function for graph editing."""

from __future__ import annotations

from . import reroute
from . import select
from . import subgraph
from . import util
from tensorflow.python.ops import array_ops as tf_array_ops

__all__ = [
    "detach_control_inputs",
    "detach_control_outputs",
    "detach_inputs",
    "detach_outputs",
    "detach",
    "connect",
    "bypass",
]


def detach_control_inputs(sgv):
    """Detach all the external control inputs of the subgraph sgv.

    Args:
      sgv: the subgraph view to be detached. This argument is converted to a
        subgraph using the same rules as the function subgraph.make_view.
    """
    sgv = subgraph.make_view(sgv)
    for op in sgv.ops:
        cops = [cop for cop in op.control_inputs if cop not in sgv.ops]
        reroute.remove_control_inputs(op, cops)


def detach_control_outputs(sgv, control_outputs):
    """Detach all the external control outputs of the subgraph sgv.

    Args:
      sgv: the subgraph view to be detached. This argument is converted to a
        subgraph using the same rules as the function subgraph.make_view.
      control_outputs: a util.ControlOutputs instance.
    """
    if not isinstance(control_outputs, util.ControlOutputs):
        raise TypeError("Expected a util.ControlOutputs, got: {}", type(control_outputs))
    control_outputs.update()
    sgv = subgraph.make_view(sgv)
    for op in sgv.ops:
        for cop in control_outputs.get(op):
            if cop not in sgv.ops:
                reroute.remove_control_inputs(cop, op)


def detach_inputs(sgv, control_inputs=False):
    """Detach the inputs of a subgraph view.

    Args:
      sgv: the subgraph view to be detached. This argument is converted to a
        subgraph using the same rules as the function subgraph.make_view.
        Note that sgv is modified in place.
      control_inputs: if True control_inputs are also detached.
    Returns:
      A tuple `(sgv, input_placeholders)` where
        `sgv` is a new subgraph view of the detached subgraph;
        `input_placeholders` is a list of the created input placeholders.
    Raises:
      StandardError: if sgv cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    sgv = subgraph.make_view(sgv)

    with sgv.graph.as_default():
        input_placeholders = [
            tf_array_ops.placeholder(dtype=input_t.dtype, name=util.placeholder_name(input_t)) for input_t in sgv.inputs
        ]

    reroute.swap_inputs(sgv, input_placeholders)
    if control_inputs:
        detach_control_inputs(sgv)
    return sgv, input_placeholders


def detach_outputs(sgv, control_outputs=None):
    """Detach the output of a subgraph view.

    Args:
      sgv: the subgraph view to be detached. This argument is converted to a
        subgraph using the same rules as the function subgraph.make_view.
        Note that sgv is modified in place.
      control_outputs: a util.ControlOutputs instance or None. If not None the
        control outputs are also detached.
    Returns:
      A tuple `(sgv, output_placeholders)` where
        `sgv` is a new subgraph view of the detached subgraph;
        `output_placeholders` is a list of the created output placeholders.
    Raises:
      StandardError: if sgv cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    sgv = subgraph.make_view(sgv)
    # only select outputs with consumers
    sgv_ = sgv.remap_outputs([output_id for output_id, output_t in enumerate(sgv.outputs) if output_t.consumers()])
    # create consumer subgraph and remap
    consumers_sgv = subgraph.SubGraphView(sgv_.consumers())
    consumers_sgv = consumers_sgv.remap_inputs(
        [input_id for input_id, input_t in enumerate(consumers_sgv.inputs) if input_t in sgv_.outputs]
    )

    with sgv_.graph.as_default():
        output_placeholders = [util.make_placeholder_from_tensor(input_t) for input_t in consumers_sgv.inputs]

    reroute.swap_outputs(sgv_, output_placeholders)
    if control_outputs is not None:
        detach_control_outputs(sgv_, control_outputs)
    return sgv_, output_placeholders


def detach(sgv, control_inputs=False, control_outputs=None, control_ios=None):
    """Detach both the inputs and the outputs of a subgraph view.

    Args:
      sgv: the subgraph view to be detached. This argument is converted to a
        subgraph using the same rules as the function subgraph.make_view.
        Note that sgv is modified in place.
      control_inputs: A boolean indicating whether control inputs are enabled.
      control_outputs: An instance of util.ControlOutputs or None. If not None,
        control outputs are enabled.
      control_ios:  An instance of util.ControlOutputs or None. If not None, both
        control inputs and control outputs are enabled. This is equivalent to set
        control_inputs to True and control_outputs to the util.ControlOutputs
        instance.
    Returns:
      A tuple `(sgv, detached_inputs, detached_outputs)` where:
      `sgv` is a new subgraph view of the detached subgraph;
      `detach_inputs` is a list of the created input placeholders;
      `detach_outputs` is a list of the created output placeholders.
    Raises:
      StandardError: if sgv cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    control_inputs, control_outputs = select.check_cios(control_inputs, control_outputs, control_ios)
    _, detached_inputs = detach_inputs(sgv, control_inputs)
    _, detached_outputs = detach_outputs(sgv, control_outputs)
    return sgv, detached_inputs, detached_outputs


def connect(sgv0, sgv1, disconnect_first=False):
    """Connect the outputs of sgv0 to the inputs of sgv1.

    Args:
      sgv0: the first subgraph to have its outputs swapped. This argument is
        converted to a subgraph using the same rules as the function
        subgraph.make_view.
        Note that sgv0 is modified in place.
      sgv1: the second subgraph to have its outputs swapped. This argument is
        converted to a subgraph using the same rules as the function
        subgraph.make_view.
        Note that sgv1 is modified in place.
      disconnect_first: if True the current outputs of sgv0 are disconnected.
    Returns:
      A tuple `(sgv0, sgv1)` of the now connected subgraphs.
    Raises:
      StandardError: if sgv0 or sgv1 cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    sgv0 = subgraph.make_view(sgv0)
    sgv1 = subgraph.make_view(sgv1)
    util.check_graphs(sgv0, sgv1)
    if disconnect_first:
        detach_outputs(sgv0)
    sgv0_outputs = subgraph.SubGraphView(passthrough_ts=sgv0.outputs)
    reroute.reroute_inputs(sgv0_outputs, sgv1)
    return sgv0, sgv1


def bypass(sgv):
    """Bypass the given subgraph by connecting its inputs to its outputs.

    Args:
      sgv: the subgraph view to be bypassed. This argument is converted to a
        subgraph using the same rules than the function subgraph.make_view.
        Note that sgv is modified in place.
    Returns:
      A tuple `(sgv, detached_inputs)` where:
        `sgv` is a new subgraph view of the bypassed subgraph;
        `detached_inputs` is a list of the created input placeholders.
    Raises:
      StandardError: if sgv cannot be converted to a SubGraphView using
        the same rules than the function subgraph.make_view.
    """
    # TODO(fkp): allows to plug sgv.inputs to individual sgv.outputs consumers
    sgv = subgraph.make_view(sgv)
    sgv_inputs = list(sgv.inputs)
    sgv, detached_inputs = detach_inputs(sgv)
    reroute.reroute_ts(sgv_inputs, sgv.outputs)
    return sgv, detached_inputs



"""
ResultWithReason class.
"""

from __future__ import annotations

from typing import TypeVar, Generic, Optional
from dataclasses import dataclass


T = TypeVar("T")


@dataclass
class ResultWithReason(Generic[T]):
    """
    This is a wrapper class for a result value, which can also have a reason.
    """

    result: T
    reason: Optional[str] = None



"""
Provides :class:`Config` and some related helpers.
"""

from __future__ import annotations

__author__ = "Patrick Doetsch"
__credits__ = ["Patrick Doetsch", "Paul Voigtlaender"]

from typing import Optional
import contextlib
import sys
import typing
import os
import types as _types


class Config:
    """
    Reads in some config file, and provides access to the key/value items.
    We support some simple text-line-based config, JSON, and Python format.
    """

    def __init__(self, items=None):
        """
        :param dict[str]|None items: optional initial typed_dict
        """
        self.dict = {}  # type: typing.Dict[str, typing.List[str]]
        self.typed_dict = {}  # :type: typing.Dict[str]  # could be loaded via JSON or so
        self.network_topology_json = None  # type: typing.Optional[str]
        self.files = []
        if items is not None:
            self.typed_dict.update(items)

    def __getstate__(self):
        import io
        from returnn.util.task_system import Pickler

        class _CustomPickler(Pickler):
            use_whichmodule = False

            # This will trigger potentially some fallback of task_system.Pickler,
            # like _save_type_fallback.
            # Also, do not use the intelligent logic for our own dict.
            # We explicitly want to pickle it as-is.
            module_name_black_list = {_PyModuleName}

        buffer = io.BytesIO()
        pickler = _CustomPickler(buffer)
        memo_idx = len(pickler.memo)
        pickler.memo[id(self)] = memo_idx, self
        pickler.dump(self.typed_dict)

        return {
            "_pid": os.getpid(),
            "_self_memo_idx": memo_idx,
            "_typed_dict_pickled": buffer.getvalue(),
            "_is_global": self is get_global_config(raise_exception=False),
        }

    def __setstate__(self, state):
        import io

        # Use pure-Python unpickling to be able to extend the memo.
        # noinspection PyUnresolvedReferences,PyProtectedMember
        from pickle import _Unpickler

        self.__init__()

        buffer = io.BytesIO(state["_typed_dict_pickled"])
        unpickler = _Unpickler(buffer)
        unpickler.memo[state["_self_memo_idx"]] = self
        self.typed_dict = unpickler.load()

        if state["_is_global"] and os.getpid() != state["_pid"]:
            set_global_config(self)
            _global_config_as_py_module_proxy_setup()

    def load_file(self, f):
        """
        Reads the configuration parameters from a file and adds them to the inner set of parameters.

        :param string|io.TextIOBase|io.StringIO f:
        """
        if isinstance(f, str):
            assert os.path.isfile(f), "config file not found: %r" % f
            self.files.append(f)
            filename = f
            dirname = os.path.dirname(filename) or "."
            content = open(filename).read()
        else:
            # assume stream-like
            filename = "<config string>"
            dirname = None
            content = f.read()
        content = content.strip()
        if content.startswith("#!") or filename.endswith(".py"):  # assume Python
            if dirname and os.path.exists(f"{dirname}/__init__.py") and filename.endswith(".py"):
                # It looks like a Python module inside a Python package.
                # Import it as a module.
                import importlib

                basedir = os.path.abspath(dirname)
                while os.path.exists(f"{basedir}/__init__.py"):
                    basedir = os.path.dirname(basedir)
                if basedir not in sys.path:
                    sys.path.insert(0, basedir)
                modname = os.path.relpath(dirname, basedir).replace("/", ".") + "." + os.path.basename(filename)[:-3]
                mod = importlib.import_module(modname)
                self.update(vars(mod))

            else:
                # Directly execute the Python code.
                from returnn.util.basic import custom_exec

                # Operate inplace on ourselves.
                # Also, we want that it's available as the globals() dict, so that defined functions behave well
                # (they would lose the local context otherwise).
                user_ns = self.typed_dict
                # Always overwrite:
                user_ns.update({"config": self, "__file__": filename, "__name__": _PyModuleName})
                custom_exec(content, filename, user_ns, user_ns)
                _global_config_as_py_module_proxy_setup()
            return
        if content.startswith("{"):  # assume JSON
            from returnn.util.basic import load_json

            json_content = load_json(content=content)
            assert isinstance(json_content, dict)
            self.update(json_content)
            return
        # old line-based format
        for line in content.splitlines():
            if "#" in line:  # Strip away comment.
                line = line[: line.index("#")]
            line = line.strip()
            if not line:
                continue
            line = line.split(None, 1)
            assert len(line) == 2, "unable to parse config line: %r" % line
            self.add_line(key=line[0], value=line[1])

    @classmethod
    def get_config_file_type(cls, f):
        """
        :param str f: file path
        :return: "py", "js" or "txt"
        :rtype: str
        """
        with open(f, "r") as f:
            start = f.read(3)
        if start.startswith("#!"):
            return "py"
        if start.startswith("{"):
            return "js"
        return "txt"

    def parse_cmd_args(self, args):
        """
        :param list[str]|tuple[str] args:
        """
        from optparse import OptionParser

        parser = OptionParser()
        parser.add_option(
            "-a",
            "--activation",
            dest="activation",
            help="[STRING/LIST] Activation functions: logistic, tanh, softsign, relu, identity, zero, one, maxout.",
        )
        parser.add_option(
            "-b",
            "--batch_size",
            dest="batch_size",
            help="[INTEGER/TUPLE] Maximal number of frames per batch (optional: shift of batching window).",
        )
        parser.add_option(
            "-c",
            "--chunking",
            dest="chunking",
            help="[INTEGER/TUPLE] Maximal number of frames per sequence (optional: shift of chunking window).",
        )
        parser.add_option("-d", "--description", dest="description", help="[STRING] Description of experiment.")
        parser.add_option("-e", "--epoch", dest="epoch", help="[INTEGER] Starting epoch.")
        parser.add_option("-E", "--eval", dest="eval", help="[STRING] eval file path")
        parser.add_option(
            "-f",
            "--gate_factors",
            dest="gate_factors",
            help="[none/local/global] Enables pooled (local) or separate (global) coefficients on gates.",
        )
        parser.add_option("-g", "--lreg", dest="lreg", help="[FLOAT] L1 or L2 regularization.")
        parser.add_option(
            "-i",
            "--save_interval",
            dest="save_interval",
            help="[INTEGER] Number of epochs until a new model will be saved.",
        )
        parser.add_option("-j", "--dropout", dest="dropout", help="[FLOAT] Dropout probability (0 to disable).")
        parser.add_option(
            "-k", "--output_file", dest="output_file", help="[STRING] Path to target file for network output."
        )
        parser.add_option("-l", "--log", dest="log", help="[STRING] Log file path.")
        parser.add_option("-L", "--load", dest="load", help="[STRING] load model file path.")
        parser.add_option(
            "-m", "--momentum", dest="momentum", help="[FLOAT] Momentum term in gradient descent optimization."
        )
        parser.add_option(
            "-n", "--num_epochs", dest="num_epochs", help="[INTEGER] Number of epochs that should be trained."
        )
        parser.add_option("-o", "--order", dest="order", help="[default/sorted/random] Ordering of sequences.")
        parser.add_option("-p", "--loss", dest="loss", help="[loglik/sse/ctc] Objective function to be optimized.")
        parser.add_option(
            "-q",
            "--cache",
            dest="cache",
            help="[INTEGER] Cache size in bytes (supports notation for kilo (K), mega (M) and gigabyte (G)).",
        )
        parser.add_option(
            "-r",
            "--learning_rate",
            dest="learning_rate",
            help="[FLOAT] Learning rate in gradient descent optimization.",
        )
        parser.add_option(
            "-s", "--hidden_sizes", dest="hidden_sizes", help="[INTEGER/LIST] Number of units in hidden layers."
        )
        parser.add_option(
            "-t",
            "--truncate",
            dest="truncate",
            help="[INTEGER] Truncates sequence in BPTT routine after specified number of timesteps (-1 to disable).",
        )
        parser.add_option(
            "-u",
            "--device",
            dest="device",
            help="[STRING/LIST] CPU and GPU devices that should be used (example: gpu0,cpu[1-6] or gpu,cpu*).",
        )
        parser.add_option("-v", "--verbose", dest="log_verbosity", help="[INTEGER] Verbosity level from 0 - 5.")
        parser.add_option("-w", "--window", dest="window", help="[INTEGER] Width of sliding window over sequence.")
        parser.add_option("-x", "--task", dest="task", help="[train/forward/analyze] Task of the current program call.")
        parser.add_option(
            "-y", "--hidden_type", dest="hidden_type", help="[VALUE/LIST] Hidden layer types: forward, recurrent, lstm."
        )
        parser.add_option(
            "-z", "--max_sequences", dest="max_seqs", help="[INTEGER] Maximal number of sequences per batch."
        )
        parser.add_option("--config", dest="load_config", help="[STRING] load config")
        (options, args) = parser.parse_args(list(args))
        options = vars(options)
        for opt in options.keys():
            if options[opt] is not None:
                if opt == "load_config":
                    self.load_file(options[opt])
                else:
                    self.add_line(opt, options[opt])
        assert len(args) % 2 == 0, "expect (++key, value) config tuples in remaining args: %r" % args
        for i in range(0, len(args), 2):
            key, value = args[i : i + 2]
            assert key[0:2] == "++", "expect key prefixed with '++' in (%r, %r)" % (key, value)
            if value[:2] == "+-":
                value = value[1:]  # otherwise we never could specify things like "++threshold -0.1"
            self.add_line(key=key[2:], value=value)

    def add_line(self, key, value):
        """
        Adds one specific configuration (key,value) pair to the inner set of parameters
        :type key: str
        :type value: str
        """
        if key in self.typed_dict:
            # This is a special case. We overwrite a config value which was typed before.
            # E.g. this could have been loaded via a Python config file.
            # We want to keep the entry in self.typed_dict because there might be functions/lambdas inside
            # the config which require the global variable to be available.
            # See :func:`test_rnn_init_config_py_global_var`.
            value_type = type(self.typed_dict[key])
            if value_type == str:
                pass  # keep as-is
            else:
                try:
                    value = eval(value)
                except SyntaxError:
                    from returnn.log import log

                    print(
                        "WARNING: can't evaluate config param %r to previous type: %s. Keeping as string."
                        % (value, value_type),
                        file=log.v1,
                    )
            self.typed_dict[key] = value
            return
        if value.find(",") > 0:
            value = value.split(",")
        else:
            value = [value]
        if key == "include":
            for f in value:
                self.load_file(f)
        else:
            self.dict[key] = value

    def has(self, key):
        """
        Returns whether the given key is present in the inner set of parameters
        :type key: string
        :rtype: boolean
        :returns True if and only if the given key is in the inner set of parameters
        """
        if key in self.typed_dict:
            return True
        return key in self.dict

    def is_typed(self, key):
        """
        :type key: string
        :rtype: boolean
        :returns True if and only if the value of the given key has a specified data type
        """
        return key in self.typed_dict

    def is_true(self, key, default=False):
        """
        :param str key:
        :param bool default:
        :return: bool(value) if it is set or default
        :rtype: bool
        """
        if self.is_typed(key):
            return bool(self.typed_dict[key])
        return self.bool(key, default=default)

    def is_of_type(self, key, types):
        """
        :param str key:
        :param type|tuple[type] types: for isinstance() check
        :return: whether is_typed(key) is True and isinstance(value, types) is True
        :rtype: bool
        """
        if key in self.typed_dict:
            return isinstance(self.typed_dict[key], types)
        return False

    def get_of_type(self, key, types, default=None):
        """
        :param str key:
        :param type|list[type]|T types: for isinstance() check
        :param T|None default:
        :return: if is_of_type(key, types) is True, returns the value, otherwise default
        :rtype: T
        """
        if self.is_of_type(key, types):
            return self.typed_dict[key]
        return default

    def set(self, key, value):
        """
        :type key: str
        :type value: list[str] | str | int | float | bool | dict | None
        """
        self.typed_dict[key] = value

    def update(self, dikt):
        """
        :type dikt: dict
        """
        for key, value in dikt.items():
            self.set(key, value)

    def _hack_value_reading_debug(self):
        orig_value_func = self.value

        def wrapped_value_func(*args, **kwargs):
            """
            Wrapped func.
            """
            res = orig_value_func(*args, **kwargs)
            print(
                "Config.value(%s) -> %r"
                % (", ".join(list(map(repr, args)) + ["%s=%r" % (k, v) for (k, v) in kwargs.items()]), res)
            )
            return res

        setattr(self, "value", wrapped_value_func)

    def value(self, key, default, index=None, list_join_str=","):
        """
        :type key: str
        :type default: T
        :type index: int | None
        :param str list_join_str:
        :rtype: str | T
        """
        if key in self.typed_dict:
            ls = self.typed_dict[key]
            if index is None:
                if isinstance(ls, (list, tuple)):
                    return list_join_str.join([str(v) for v in ls])
                elif ls is None:
                    return default
                else:
                    return str(ls)
            else:
                return str(ls[index])
        if key in self.dict:
            ls = self.dict[key]
            if index is None:
                return list_join_str.join(ls)
            else:
                return ls[index]
        return default

    def typed_value(self, key, default=None, index=None):
        """
        :type key: str
        :type default: T
        :type index: int | None
        :rtype: T | typing.Any
        """
        value = self.typed_dict.get(key, default)
        if index is not None:
            assert isinstance(index, int)
            if isinstance(value, (list, tuple)):
                value = value[index]
            else:
                assert index == 0
        return value

    def opt_typed_value(self, key, default=None):
        """
        :param str key:
        :param T|None default:
        :rtype: T|object|str|None
        """
        if key in self.typed_dict:
            return self.typed_dict[key]
        return self.value(key, default)

    def int(self, key, default, index=0):
        """
        Parses the value of the given key as integer, returning default if not existent
        :type key: str
        :type default: T
        :type index: int
        :rtype: int | T
        """
        if key in self.typed_dict:
            value = self.typed_value(key, default=default, index=index)
            if value is not None:
                assert isinstance(value, int)
            return value
        if key in self.dict:
            return int(self.value(key, default, index))
        return default

    def bool(self, key, default, index=0):
        """
        Parses the value of the given key as boolean, returning default if not existent
        :type key: str
        :type default: T
        :type index: bool
        :rtype: bool | T
        """
        if key in self.typed_dict:
            value = self.typed_value(key, default=default, index=index)
            if isinstance(value, int):
                value = bool(value)
            if value is not None:
                assert isinstance(value, bool)
            return value
        if key not in self.dict:
            return default
        v = str(self.value(key, None, index))
        if not v:
            return default
        from returnn.util.basic import to_bool

        return to_bool(v)

    def bool_or_other(self, key, default, index=0):
        """
        :param str key:
        :param T default:
        :param int index:
        :return: if we have typed value, just as-is. otherwise try to convert to bool. or default if not there.
        :rtype: bool|T|object
        """
        if key in self.typed_dict:
            return self.typed_value(key, default=default, index=index)
        if key not in self.dict:
            return default
        v = str(self.value(key, None, index))
        if not v:
            return default
        from returnn.util.basic import to_bool

        try:
            return to_bool(v)
        except ValueError:
            return v

    def float(self, key, default, index=0):
        """
        Parses the value of the given key as float, returning default if not existent
        :type key: str
        :type default: T
        :type index: int
        :rtype: float | T
        """
        if key in self.typed_dict:
            value = self.typed_value(key, default=default, index=index)
        else:
            value = self.value(key, default, index)
        if value is not None:
            if isinstance(value, str):
                # Special case for float as str. We automatically cast this case.
                # This is also to handle special values such as "inf".
                value = float(value)
            assert isinstance(value, (int, float))
        return value

    def list(self, key, default=None):
        """
        :type key: str
        :type default: T
        :rtype: list[str] | T
        """
        if default is None:
            default = []
        if key in self.typed_dict:
            value = self.typed_value(key, default=default)
            if value is None:
                return default
            if not isinstance(value, (tuple, list)):
                value = [value]
            return list(value)
        if key not in self.dict:
            return default
        return self.dict[key]

    def int_list(self, key, default=None):
        """
        :type key: str
        :type default: T
        :rtype: list[int] | T
        """
        if default is None:
            default = []
        if key in self.typed_dict:
            value = self.typed_value(key, default=default)
            if value is None:
                return default
            if not isinstance(value, (tuple, list)):
                value = [value]
            for x in value:
                assert isinstance(x, int)
            return list(value)
        return [int(x) for x in self.list(key, default)]

    def float_list(self, key, default=None):
        """
        :type key: str
        :type default: T
        :rtype: list[float] | T
        """
        if default is None:
            default = []
        if key in self.typed_dict:
            value = self.typed_value(key, default=default)
            if value is None:
                return default
            if not isinstance(value, (tuple, list)):
                value = [value]
            for x in value:
                assert isinstance(x, (float, int))
            return list(value)
        return [float(x) for x in self.list(key, default)]

    def int_pair(self, key, default=None):
        """
        :param str key:
        :param (int,int)|None default:
        :rtype: (int,int)
        """
        if default is None:
            default = (0, 0)
        if not self.has(key):
            return default
        if key in self.typed_dict:
            value = self.typed_value(key, default=default)
            if not isinstance(value, (tuple, list)):
                value = (value, value)
            assert len(value) == 2
            for x in value:
                assert isinstance(x, int)
            return tuple(value)
        value = self.value(key, "")
        if ":" in value:
            return int(value.split(":")[0]), int(value.split(":")[1])
        else:
            return int(value), int(value)


_global_config: Optional[Config] = None


@contextlib.contextmanager
def global_config_ctx(config: Optional[Config]):
    """
    sets the config as global config in this context,
    and recovers the original global config afterwards
    """
    global _global_config
    prev_global_config = _global_config
    try:
        set_global_config(config)
        yield
    finally:
        _global_config = prev_global_config


def set_global_config(config: Optional[Config]):
    """
    Will define the global config, returned by :func:`get_global_config`

    :param config:
    """
    _get_or_set_config_via_tf_default_graph(config, assign=True)
    global _global_config
    _global_config = config


def get_global_config(*, raise_exception: bool = True, auto_create: bool = False, return_empty_if_none: bool = False):
    """
    :param raise_exception: if no global config is found, raise an exception, otherwise return None
    :param auto_create: if no global config is found, it creates one, registers it as global, and returns it
    :param return_empty_if_none: if no global config is found, it creates one (which is empty) and returns it
    :rtype: Config|None
    """
    config = _get_or_set_config_via_tf_default_graph()
    if config:
        return config
    if _global_config:
        return _global_config
    # We are the main process.
    import sys

    main_mod = sys.modules["__main__"]  # should be rnn.py
    if hasattr(main_mod, "config") and isinstance(main_mod.config, Config):
        return main_mod.config
    # Maybe __main__ is not rnn.py, or config not yet loaded.
    # Anyway, try directly. (E.g. for SprintInterface.)
    import returnn.__main__ as rnn

    if isinstance(rnn.config, Config):
        return rnn.config
    if auto_create:
        config = Config()
        set_global_config(config)
        return config
    if return_empty_if_none:
        return Config()
    if raise_exception:
        raise Exception("No global config found.")
    return None


def _get_or_set_config_via_tf_default_graph(config: Optional[Config] = None, assign: bool = False):
    """
    This is done in a safe way, and might just be a no-op.
    When TF is not imported yet, it will just return.

    :param config: if set, will set it
    :param assign: whether to set it (even if config=None)
    :rtype: Config|None
    """
    if "tensorflow" not in sys.modules:
        return None
    from returnn.tf.compat import v1 as tf_v1

    graph = tf_v1.get_default_graph()
    # We could use collection refs, but this could cause other problems,
    # and is more complicated than what we need.
    # We just use a custom own attrib.
    attrib_name = "_RETURNN_config_in_graph"
    if assign or config:
        setattr(graph, attrib_name, config)
    return getattr(graph, attrib_name, None)


def network_json_from_config(config):
    """
    :param Config config:
    :rtype: dict[str]
    """
    if config.has("network") and config.is_typed("network"):
        json_content = config.typed_value("network")
        assert isinstance(json_content, dict)
        assert json_content
        return json_content
    else:
        raise ValueError("Network is not defined in config. Define `network`.")


def tf_should_use_gpu(config):
    """
    :param Config config:
    :rtype: bool
    """
    cfg_dev = config.value("device", None)
    # Short path.
    if cfg_dev == "gpu":
        return True
    if cfg_dev == "cpu":
        return False
    if not cfg_dev:
        # Better default: Use GPU if available.
        from returnn.log import log
        from returnn.tf.util.basic import is_gpu_available

        if is_gpu_available():
            print("Device not set explicitly, and we found a GPU, which we will use.", file=log.v2)
            config.set("device", "gpu")
            return True
        else:
            print("Device not set explicitly, and no GPU found.", file=log.v2)
            config.set("device", "cpu")
            return False
    else:
        raise ValueError("Currently unsupported TF device %r specified" % (cfg_dev,))


_PyModuleName = "__returnn_config__"


def _global_config_as_py_module_proxy_setup():
    if _PyModuleName in sys.modules:
        return
    sys.modules[_PyModuleName] = _GlobalConfigAsPyModuleProxy(_PyModuleName)


class _GlobalConfigAsPyModuleProxy(_types.ModuleType):
    """
    Takes :func:`get_global_config`, and makes its ``typed_dict`` available as module attributes.
    """

    @staticmethod
    def _get_config() -> Optional[Config]:
        """
        :return: config or None if not available anymore
        """
        return get_global_config(raise_exception=False)

    def __getattribute__(self, item):
        # Implement also __getattribute__ such that early access to just self.__dict__ (e.g. via vars(self)) also works.
        if item == "__dict__":
            cfg: Optional[Config] = self._get_config()
            if not cfg:
                return {}
            return cfg
        return super().__getattribute__(item)

    def __getattr__(self, item):
        cfg: Optional[Config] = self._get_config()
        if not cfg:
            raise AttributeError("config %s not loaded anymore" % self.__name__)
        if item not in cfg.typed_dict:
            raise AttributeError("config %s has no attribute %r" % (self.__name__, item))
        return cfg.typed_dict[item]

    def __dir__(self):
        cfg: Optional[Config] = self._get_config()
        if not cfg:
            return []
        return sorted(cfg.typed_dict.keys())

    def __setattr__(self, key, value):
        if key in ["__file__"]:
            super().__setattr__(key, value)
            return
        cfg: Optional[Config] = self._get_config()
        if cfg:
            cfg.typed_dict[key] = value



#!/usr/bin/env python
# coding=utf-8
import h5py
import numpy
from scipy.ndimage import imread
from scipy.misc import imsave, imresize
import os
import glob
import sys
import errno
import re


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def hdf5_strings(handle, name, data):
    try:
        S = max([len(d) for d in data])
        dset = handle.create_dataset(name, (len(data),), dtype="S" + str(S))
        dset[...] = data
    except Exception:
        dt = h5py.special_dtype(vlen=str)
        del handle[name]
        dset = handle.create_dataset(name, (len(data),), dtype=dt)
        dset[...] = data


def load_char_list(char_list_path):
    charlist = []
    with open(char_list_path) as f:
        for l in f:
            charlist.append(l.strip())
    return charlist


def load_file_list_and_transcriptions_and_sizes_and_n_labels(file_list_path, char_list_path, pad_whitespace, base_path):
    charlist = load_char_list(char_list_path)
    file_list = []
    transcription_list = []
    size_list = []
    with open(file_list_path) as f:
        for l in f:
            # IAM format
            if l.startswith("#"):
                continue
            sp = l.split()
            status = sp[1]
            assert status in ("ok", "err"), status
            assert len(sp) >= 9, l
            name = sp[0]
            text = "".join(sp[8:])
            # add space before '
            text = re.sub("([^|])'", "\g<1>|'", text)
            width = int(sp[6])
            height = int(sp[7])
            if height < 1 or width < 1:
                continue
            size_list.append((height, width))
            s = name.split("-")
            name = base_path + name + ".png"
            text = [charlist.index(c) for c in text]
            if pad_whitespace:
                text = [charlist.index("|")] + text + [charlist.index("|")]
            file_list.append(name)
            transcription_list.append(text)
    return file_list, transcription_list, size_list, len(charlist)


def write_to_hdf(
    file_list, transcription_list, charlist, n_labels, out_file_name, dataset_prefix, pad_y=15, pad_x=15, compress=True
):
    with h5py.File(out_file_name, "w") as f:
        f.attrs["inputPattSize"] = 1
        f.attrs["numDims"] = 1
        f.attrs["numSeqs"] = len(file_list)
        classes = charlist

        inputs = []
        sizes = []
        seq_lengths = []
        targets = []
        for i, (img_name, transcription) in enumerate(zip(file_list, transcription_list)):
            targets += transcription
            img = imread(img_name)
            img = 255 - img
            img = numpy.pad(img, ((pad_y, pad_y), (pad_x, pad_x)), "constant")
            sizes.append(img.shape)
            img = img.reshape(img.size, 1)
            inputs.append(img)
            seq_lengths.append([[img.size, len(transcription), 2]])
            if i % 100 == 0:
                print(i, "/", len(file_list))

        inputs = numpy.concatenate(inputs, axis=0)
        sizes = numpy.concatenate(numpy.array(sizes, dtype="int32"), axis=0)
        seq_lengths = numpy.concatenate(numpy.array(seq_lengths, dtype="int32"), axis=0)
        targets = numpy.array(targets, dtype="int32")

        f.attrs["numTimesteps"] = inputs.shape[0]

        if compress:
            f.create_dataset("inputs", compression="gzip", data=inputs.astype("float32") / 255.0)
        else:
            f["inputs"] = inputs.astype("float32") / 255.0
        hdf5_strings(f, "labels", classes)
        f["seqLengths"] = seq_lengths
        seq_tags = [dataset_prefix + "/" + tag.split("/")[-1].split(".png")[0] for tag in file_list]
        hdf5_strings(f, "seqTags", seq_tags)

        f["targets/data/classes"] = targets
        f["targets/data/sizes"] = sizes
        hdf5_strings(f, "targets/labels/classes", classes)
        hdf5_strings(f, "targets/labels/sizes", ["foo"])  # TODO, can we just omit it?
        g = f.create_group("targets/size")
        g.attrs["classes"] = len(classes)
        g.attrs["sizes"] = 2


def sort_by_size(file_list, transcription_list, size_list):
    zipped = list(zip(file_list, transcription_list, size_list))
    # sort by (width, height)
    sorted_lists = sorted(zipped, key=lambda x: (x[2][1], x[2][0]))
    return [x[0] for x in sorted_lists], [x[1] for x in sorted_lists], [x[2] for x in sorted_lists]


def convert(
    file_list_path, char_list_path, selections, out_file_names, pad_whitespace, dataset_prefix, base_path, compress
):
    charlist = load_char_list(char_list_path)
    file_list, transcription_list, size_list, n_labels = load_file_list_and_transcriptions_and_sizes_and_n_labels(
        file_list_path, char_list_path, pad_whitespace, base_path
    )
    file_list, transcription_list, size_list = sort_by_size(file_list, transcription_list, size_list)

    for selection, out_file_name in zip(selections, out_file_names):
        print(out_file_name)
        selection_set = set(selection)
        assert selection_set.issubset(set(x.split("/")[-1].split(".png")[0] for x in file_list))
        selected_file_list = []
        selected_transcription_list = []
        for f, t in zip(file_list, transcription_list):
            if f.split("/")[-1].split(".png")[0] in selection_set:
                selected_file_list.append(f)
                selected_transcription_list.append(t)
        write_to_hdf(
            selected_file_list,
            selected_transcription_list,
            charlist,
            n_labels,
            out_file_name,
            dataset_prefix,
            compress=compress,
        )


def get_image_list(train_list_path):
    with open(train_list_path) as f:
        imgs = f.readlines()
    imgs = [img.replace("\n", "") for img in imgs]
    return imgs


def get_train_and_train_valid_lists(train_list_path, blacklist, train_fraction=0.9):
    with open(train_list_path) as f:
        imgs = f.readlines()
    imgs = [img.replace("\n", "") for img in imgs]
    n_train = int(round(train_fraction * len(imgs)))

    train_imgs = imgs[:n_train]
    train_valid_imgs = imgs[n_train:]

    n_before = len(train_imgs)
    train_imgs = [s for s in train_imgs if s not in blacklist]
    n_after = len(train_imgs)
    if n_before != n_after:
        print("removed", n_before - n_after, "blacklisted images from train")

    n_before = len(train_valid_imgs)
    train_valid_imgs = [s for s in train_valid_imgs if s not in blacklist]
    n_after = len(train_valid_imgs)
    if n_before != n_after:
        print("removed", n_before - n_after, "blacklisted images from train_valid")

    return train_imgs, train_valid_imgs


def convert_IAM_lines_demo(base_path_imgs, tag, blacklist=[]):
    base_path_out = "features/" + tag + "/"
    mkdir_p(base_path_out)
    file_list_path = "lines.txt"
    char_list_path = "chars.txt"
    selection_list_path = "split/demo.txt"
    out_file_name_demo = base_path_out + "demo.h5"

    print("converting IAM_lines to", out_file_name_demo)
    demo_list = ["a01-000u-00", "a01-007-04", "a01-007-06"]
    selections = [demo_list]
    out_file_names = [out_file_name_demo]

    convert(
        file_list_path,
        char_list_path,
        selections,
        out_file_names,
        pad_whitespace=True,
        dataset_prefix="trainset",
        base_path=base_path_imgs,
        compress=True,
    )


def convert_IAM_lines_train(base_path_imgs, tag, blacklist=[]):
    base_path_out = "features/" + tag + "/"
    mkdir_p(base_path_out)
    file_list_path = "lines.txt"
    char_list_path = "chars.txt"
    selection_list_path = "split/train.txt"
    out_file_name_train1 = base_path_out + "train.1.h5"
    out_file_name_train2 = base_path_out + "train.2.h5"
    out_file_name_train_valid = base_path_out + "train_valid.h5"

    print("converting IAM_lines to", out_file_name_train1, "and", out_file_name_train2)
    train_list, train_valid_list = get_train_and_train_valid_lists(selection_list_path, blacklist, 0.9)
    len1 = len(train_list) / 2
    train_list1 = train_list[:len1]
    train_list2 = train_list[len1:]
    selections = [train_list1, train_list2, train_valid_list]
    out_file_names = [out_file_name_train1, out_file_name_train2, out_file_name_train_valid]

    convert(
        file_list_path,
        char_list_path,
        selections,
        out_file_names,
        pad_whitespace=True,
        dataset_prefix="trainset",
        base_path=base_path_imgs,
        compress=True,
    )


def convert_IAM_lines_valid_test(base_path_imgs, tag):
    base_path_out = "features/" + tag + "/"
    mkdir_p(base_path_out)
    char_list_path = "chars.txt"
    selection_list_path_valid = "split/valid.txt"
    selection_list_path_test = "split/eval.txt"
    out_file_name_valid = base_path_out + "valid.h5"
    out_file_name_test = base_path_out + "test.h5"
    prefix_valid = "validationset"
    prefix_test = "testset"

    charlist = load_char_list(char_list_path)
    n_labels = len(charlist)

    for selection_list_path, out_file_name, prefix in zip(
        [selection_list_path_valid, selection_list_path_test],
        [out_file_name_valid, out_file_name_test],
        [prefix_valid, prefix_test],
    ):
        selection = [x.strip() for x in open(selection_list_path).readlines()]
        imgs = []
        for sel in selection:
            pattern = base_path_imgs + sel + "-[0-9][0-9].png"
            new_imgs = glob.glob(pattern)
            assert len(new_imgs) > 0, (sel, pattern)
            imgs.extend(new_imgs)
        imgs = sorted(imgs)
        transcriptions = [[]] * len(imgs)

        print("converting IAM_lines to", out_file_name)
        write_to_hdf(imgs, transcriptions, charlist, n_labels, out_file_name, dataset_prefix=prefix, compress=True)


def main():
    base_path_imgs = "IAM_lines"
    tag = "raw"

    if base_path_imgs[-1] != "/":
        base_path_imgs += "/"
    convert_IAM_lines_demo(base_path_imgs, tag)
    # convert_IAM_lines_train(base_path_imgs, tag)
    # convert_IAM_lines_valid_test(base_path_imgs, tag)


if __name__ == "__main__":
    main()



"""
Conditional execution.

Why did we choose this API?
To allow for both eager-based and graph-based frameworks,
and also to avoid any magic happening here.
https://github.com/rwth-i6/returnn/issues/1282
"""

from __future__ import annotations
from typing import Union, TypeVar, Callable
from returnn.tensor import Tensor


T = TypeVar("T")


def cond(pred: Union[bool, Tensor], true_fn: Callable[[], T], false_fn: Callable[[], T]) -> T:
    """
    :param pred:
    :param true_fn:
    :param false_fn:
    :return: true_fn() if pred else false_fn()
    """
    if isinstance(pred, bool):
        if pred:
            return true_fn()
        else:
            return false_fn()
    assert isinstance(pred, Tensor) and pred.dims == () and pred.dtype == "bool"
    # noinspection PyProtectedMember
    backend = pred._raw_backend
    if backend.executing_eagerly():
        if bool(pred.raw_tensor):
            return true_fn()
        else:
            return false_fn()
    return backend.cond(pred, true_fn, false_fn)



"""
Control flow context
"""

from __future__ import annotations
from typing import Optional, TYPE_CHECKING, List
import returnn.util.basic as util

if TYPE_CHECKING:
    from tensorflow.python.ops.control_flow_ops import ControlFlowContext as _TFControlFlowCtx
    from .dim import Dim


__all__ = ["ControlFlowContext"]


class ControlFlowContext:
    """
    This represents the current control flow context, e.g. whether this runs in a loop or a conditional branch.

    In case of TF,
    this is a simple wrapper around the TF ControlFlowContext which comes from tf.while_loop or tf.cond.

    We have this wrapper to refer to a context which might not exist yet (e.g. at template construction time).
    Also, we might want to store additional information, such the spatial dim tag of the loop.
    """

    class Types:
        """
        Possible types of context.
        """

        Loop = "loop"
        CondTrue = "cond-true"
        CondFalse = "cond-false"

    def __init__(self, *, kind: str, identifier: str, outer_ctx: Optional[ControlFlowContext] = None):
        """
        :param kind: from ControlFlowContext.Types
        :param outer_ctx:
        """
        self.kind = kind
        self.identifier = identifier
        self._outer_ctx = outer_ctx
        self._tf_control_flow_ctx = None  # type: Optional[_TFControlFlowCtx]
        self._loop_spatial_dim = None  # type: Optional[Dim]

    def __repr__(self):
        return "ControlFlowContext{%s}" % self.repr_inner()

    def repr_inner(self):
        """
        :rtype: str
        """
        return "/".join(ctx._repr_single() for ctx in self._abs_ctx_stack())

    def _repr_single(self):
        """
        :rtype: str
        """
        s = self.kind
        if self.is_loop() and self.loop_spatial_dim:
            try:
                with util.guard_infinite_recursion(ControlFlowContext._repr_single, self):
                    s += "(%s)" % self.loop_spatial_dim.short_repr()
            except util.InfiniteRecursionDetected as exc:
                # repr should always return and not throw errors
                s += "(%s for loop_spatial_dim?)" % exc
        return s

    def _abs_ctx_stack(self):
        """
        :rtype: list[ControlFlowContext]
        :return: chain of ctx, last is self
        """
        chain = []
        ctx = self
        while ctx:
            chain.append(ctx)
            ctx = ctx.outer_ctx
        chain.reverse()
        return chain

    @classmethod
    def abs_ctx_stack(cls, ctx):
        """
        :param ControlFlowContext|None ctx:
        :rtype: list[ControlFlowContext]
        """
        if ctx:
            return ctx._abs_ctx_stack()
        return []

    @classmethod
    def abs_ctx_stack_with_root(cls, ctx):
        """
        :param ControlFlowContext|None ctx:
        :rtype: list[ControlFlowContext|None]
        :return: chain of ctx, last is self, first is None
        """
        ls = [None]  # type: List[Optional[ControlFlowContext]]
        if ctx:
            ls += ctx._abs_ctx_stack()
        return ls

    @classmethod
    def is_parent_or_same(cls, parent, child):
        """
        :param ControlFlowContext|None parent:
        :param ControlFlowContext|None child:
        :rtype: bool
        """
        if parent == child:
            return True
        if not parent:
            return True  # parent is root
        if not child:
            return False  # child is root but parent is not
        while child:
            if child == parent:
                return True
            child = child.outer_ctx
        return False

    @classmethod
    def collect_parent_dims(cls, ctx):
        """
        :param ControlFlowContext|None ctx:
        :rtype: list[Dim]
        """
        dims = []
        for ctx_ in ControlFlowContext.abs_ctx_stack(ctx):
            if ctx_.is_loop() and ctx_.loop_spatial_dim:
                dims.append(ctx_.loop_spatial_dim)
        return dims

    def is_loop(self):
        """
        :rtype: bool
        """
        return self.kind == self.Types.Loop

    def is_cond(self):
        """
        :rtype: bool
        """
        return self.kind in {self.Types.CondTrue, self.Types.CondFalse}

    @property
    def outer_ctx(self):
        """
        :rtype: ControlFlowContext|None
        """
        return self._outer_ctx

    @property
    def tf_control_flow_ctx(self):
        """
        :rtype: tensorflow.python.ops.control_flow_ops.ControlFlowContext|None
        """
        return self._tf_control_flow_ctx

    @tf_control_flow_ctx.setter
    def tf_control_flow_ctx(self, ctx):
        """
        :param tensorflow.python.ops.control_flow_ops.ControlFlowContext ctx:
        """
        if self.is_loop():
            assert ctx.IsWhileContext()
        if self.is_cond():
            assert ctx.IsCondContext()
        self._tf_control_flow_ctx = ctx

    @property
    def loop_spatial_dim(self):
        """
        :rtype: Dim|None
        """
        assert self.is_loop()
        return self._loop_spatial_dim

    @loop_spatial_dim.setter
    def loop_spatial_dim(self, dim):
        """
        :param Dim dim:
        """
        assert self.is_loop()
        self._loop_spatial_dim = dim



"""
Tests for PyTorch engine.
"""

from __future__ import annotations
import _setup_test_env  # noqa
import sys
import unittest
import tempfile
import numpy
import torch

from returnn.util import better_exchook
from returnn.config import Config, global_config_ctx
from returnn.tensor import TensorDict, Tensor
from returnn.torch.engine import Engine
from returnn.torch.updater import Updater
import returnn.frontend as rf
from returnn.forward_iface import ForwardCallbackIface
from returnn.datasets import init_dataset


def test_torch_engine_train():
    class _Model(torch.nn.Module):
        def __init__(self, **_kwargs):
            super(_Model, self).__init__()
            self.lin = torch.nn.Linear(9, 2)

        def __call__(self, x: torch.Tensor) -> torch.Tensor:
            """
            :param x: [B,T,D]
            :return: [B,T,D']
            """
            x = self.lin(x)
            return torch.nn.functional.log_softmax(x, dim=-1)

        @classmethod
        def train_step(cls, *, model: _Model, extern_data: TensorDict, **_kwargs):
            """train step"""
            data: Tensor = extern_data["data"]
            logits = model(data.raw_tensor)
            logits_packed = torch.nn.utils.rnn.pack_padded_sequence(
                logits, data.dims[1].dyn_size_ext.raw_tensor, batch_first=True, enforce_sorted=False
            )
            targets = extern_data["classes"]
            targets_packed = torch.nn.utils.rnn.pack_padded_sequence(
                targets.raw_tensor, data.dims[1].dyn_size_ext.raw_tensor, batch_first=True, enforce_sorted=False
            )
            loss = torch.nn.CrossEntropyLoss(reduction="none")(logits_packed.data, targets_packed.data.long())
            rf.get_run_ctx().mark_as_loss(name="ce", loss=loss)
            frame_error = torch.argmax(logits_packed.data, dim=-1).not_equal(targets_packed.data)
            rf.get_run_ctx().mark_as_loss(name="fer", loss=frame_error, as_error=True)

    config = Config(
        dict(
            task="train",
            device="cpu",
            extern_data={"data": {"dim": 9}, "classes": {"dim": 2, "sparse": True}},
            get_model=_Model,
            train_step=_Model.train_step,
            batch_size=500,
            optimizer={"class": "adam"},
        )
    )
    dataset = init_dataset({"class": "Task12AXDataset", "num_seqs": 100, "name": "train"})
    dataset.init_seq_order(epoch=1)

    with global_config_ctx(config):
        engine = Engine(config=config)
        engine.init_train_from_config(train_data=dataset)
        engine.train()


def test_torch_engine_forward_simple():
    def _get_model(**_kwargs):
        return torch.nn.Module()

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        rf.get_run_ctx().mark_as_default_output(extern_data["data"])

    config = Config(
        dict(
            task="forward",
            extern_data={"data": {"dim": 9}},
            batch_size=500,
            get_model=_get_model,
            forward_step=_forward_step,
        )
    )
    dataset = init_dataset({"class": "Task12AXDataset", "num_seqs": 100, "name": "dev", "fixed_random_seed": 1})
    dataset.init_seq_order(epoch=1)
    callback = ForwardCallbackIface()

    with global_config_ctx(config):
        engine = Engine(config=config)
        engine.init_network_from_config()
        engine.forward_with_callback(callback=callback, dataset=dataset)


def test_torch_engine_forward():
    def _get_model(**_kwargs):
        return torch.nn.Module()

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        rf.get_run_ctx().mark_as_default_output(extern_data["data"])

    class _ForwardCallback(ForwardCallbackIface):
        def __init__(self):
            self.num_seqs = 0
            self.init_called = False
            self.finish_called = False

        def init(self, *, model):
            assert isinstance(model, torch.nn.Module)
            assert self.num_seqs == 0
            self.init_called = True

        def process_seq(self, *, seq_tag: str, outputs: TensorDict):
            assert isinstance(seq_tag, str) and seq_tag.startswith("seq-")
            assert isinstance(outputs, TensorDict)
            out = outputs["output"]
            assert isinstance(out, Tensor)
            assert out.batch_ndim == 2 and out.batch_shape[-1] == 9
            self.num_seqs += 1

        def finish(self):
            self.finish_called = True

    config = Config(
        dict(
            task="forward",
            extern_data={"data": {"dim": 9}},
            batch_size=500,
            get_model=_get_model,
            forward_step=_forward_step,
        )
    )
    dataset = init_dataset({"class": "Task12AXDataset", "num_seqs": 100, "name": "dev", "fixed_random_seed": 1})
    dataset.init_seq_order(epoch=1)
    callback = _ForwardCallback()

    with global_config_ctx(config):
        engine = Engine(config=config)
        engine.init_network_from_config()
        engine.forward_with_callback(callback=callback, dataset=dataset)
        assert callback.num_seqs == 100
        assert callback.init_called and callback.finish_called


def test_torch_engine_forward_pure_torch_no_model_out():
    # https://github.com/rwth-i6/returnn/issues/1385
    # Automatically assume that we have batch-dim first in mark_as_output with raw tensor.
    def _get_model(**_kwargs):
        return torch.nn.Module()

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        rf.get_run_ctx().mark_as_default_output(extern_data["data"].raw_tensor)

    config = Config(
        dict(
            task="forward",
            extern_data={"data": {"dim": 9}},
            batch_size=500,
            get_model=_get_model,
            forward_step=_forward_step,
        )
    )
    dataset = init_dataset({"class": "Task12AXDataset", "num_seqs": 100, "name": "dev", "fixed_random_seed": 1})
    dataset.init_seq_order(epoch=1)
    callback = ForwardCallbackIface()

    with global_config_ctx(config):
        engine = Engine(config=config)
        engine.init_network_from_config()
        engine.forward_with_callback(callback=callback, dataset=dataset)


def test_torch_forward_raw_strings():
    # In OggZipDataset, but maybe also other datasets,
    # in combination with forward task, we get all kind of different string formats:
    #   - seq_tag (numpy U8 str)
    #   - raw (numpy object -> str)
    #   - orth (uint8 bytes)
    # Test all of them.

    from test_Dataset import create_ogg_zip_txt_only_dataset

    def _get_model(**_kwargs):
        return torch.nn.Module()

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        for key, value in extern_data.data.items():
            rf.get_run_ctx().mark_as_output(value, key)

    config = Config(
        dict(
            task="forward",
            extern_data={
                "classes": {"shape": (None,), "dim": 29, "sparse": True},
                "orth": {"shape": (None,), "dim": 256, "sparse": True},
                "raw": {"shape": (), "dtype": "string"},
            },
            batch_size=500,
            get_model=_get_model,
            forward_step=_forward_step,
        )
    )

    _demo_txt = "hello world"
    _demo_seq_tag = "seq-000000"

    class _ForwardCallback(ForwardCallbackIface):
        def process_seq(self, *, seq_tag: str, outputs: TensorDict):
            assert isinstance(seq_tag, str) and seq_tag == _demo_seq_tag
            raw = outputs["raw"].raw_tensor
            orth = outputs["orth"].raw_tensor
            classes = outputs["classes"].raw_tensor
            assert isinstance(raw, numpy.ndarray) and raw.dtype.name.startswith("str") and raw.shape == ()
            raw_ = raw.item()
            assert isinstance(raw_, str) and raw_ == _demo_txt
            assert isinstance(orth, numpy.ndarray) and orth.dtype == numpy.uint8 and orth.ndim == 1
            orth_ = orth.tostring()
            assert orth_.decode("utf8") == _demo_txt
            assert isinstance(classes, numpy.ndarray) and classes.dtype == numpy.int32 and classes.ndim == 1
            classes_ = "".join([dataset.targets.id_to_label(c) for c in classes])
            assert classes_ == _demo_txt + "."

    with global_config_ctx(config), create_ogg_zip_txt_only_dataset(text=_demo_txt, seq_tag=_demo_seq_tag) as dataset:
        dataset.init_seq_order(epoch=1)
        engine = Engine(config=config)
        engine.init_network_from_config()
        engine.forward_with_callback(callback=_ForwardCallback(), dataset=dataset)


def test_forward_beam_seq_lens():
    from returnn.tensor import Dim, batch_dim

    def _get_model(**_kwargs):
        return torch.nn.Module()

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        data = extern_data["data"]  # [batch, time, dim]
        assert data.dims[0] == batch_dim
        time_dim = data.dims[1]
        feat_dim = data.dims[2]
        beam_dim = Dim(dimension=5, name="beam")
        with rf.set_default_device_ctx(time_dim.dyn_size_ext.device):
            ext_seq_lens = rf.relu(
                rf.combine_bc(
                    time_dim.dyn_size_ext, "-", rf.range_over_dim(beam_dim, dtype=time_dim.dyn_size_ext.dtype)
                )
            )
        assert set(ext_seq_lens.dims) == {batch_dim, beam_dim}
        ext_time_dim = Dim(ext_seq_lens, name="time_with_beam")
        ext_data = rf.expand_dim(data, beam_dim)
        ext_data, _ = rf.replace_dim(ext_data, in_dim=time_dim, out_dim=ext_time_dim)
        assert set(ext_data.dims) == {batch_dim, beam_dim, ext_time_dim, feat_dim}
        rf.get_run_ctx().mark_as_output(ext_data, "ext_data", dims=(batch_dim, beam_dim, ext_time_dim, feat_dim))

    max_sizes = set()

    class _ForwardCallback(ForwardCallbackIface):
        def process_seq(self, *, seq_tag: str, outputs: TensorDict):
            out: Tensor = outputs["ext_data"]
            beam_dim, ext_time_dim, feat_dim = out.dims
            assert isinstance(ext_time_dim.dyn_size_ext.raw_tensor, numpy.ndarray)
            assert ext_time_dim.dyn_size_ext.dims == (beam_dim,)
            max_size = max(ext_time_dim.dyn_size_ext.raw_tensor)
            assert set(ext_time_dim.dyn_size_ext.raw_tensor) == set(
                range(max(max_size - beam_dim.dimension + 1, 0), max_size + 1)
            )
            max_sizes.add(max_size)

    config = Config(
        dict(
            task="forward",
            batch_size=500,
            extern_data={"data": {"dim": 9}},
            get_model=_get_model,
            forward_step=_forward_step,
        )
    )
    dataset = init_dataset({"class": "Task12AXDataset", "num_seqs": 100, "name": "dev", "fixed_random_seed": 1})
    callback = _ForwardCallback()

    with global_config_ctx(config):
        dataset.init_seq_order(epoch=1)
        engine = Engine(config=config)
        engine.init_network_from_config()
        engine.forward_with_callback(callback=callback, dataset=dataset)
        assert len(max_sizes) > 1


def test_min_seq_len():

    from returnn.datasets.generating import DummyDataset

    config = Config({"min_seq_length": 2, "batch_size": 3})
    dataset = DummyDataset(input_dim=1, output_dim=4, num_seqs=1, seq_len=1)
    dataset.initialize()
    dataset.init_seq_order(epoch=1)
    engine = Engine(config=config)
    data_loader = engine._create_data_loader(dataset)
    for _ in data_loader:
        assert False, "Should not contain sequences"

    config = Config(dict(batch_size=3))
    dataset = DummyDataset(input_dim=1, output_dim=4, num_seqs=1, seq_len=3)
    dataset.initialize()
    dataset.init_seq_order(epoch=1)
    engine = Engine(config=config)
    data_loader = engine._create_data_loader(dataset)
    for _ in data_loader:
        return
    assert False, "Should have contained sequences"


def test_max_seq_len():

    from returnn.datasets.generating import DummyDataset

    config = Config({"max_seq_length": 4, "batch_size": 3})
    dataset = DummyDataset(input_dim=1, output_dim=4, num_seqs=1, seq_len=5)
    dataset.initialize()
    dataset.init_seq_order(epoch=1)
    engine = Engine(config=config)
    data_loader = engine._create_data_loader(dataset)
    for _ in data_loader:
        assert False, "Should not contain sequences"

    config = Config(dict(batch_size=3))
    dataset = DummyDataset(input_dim=1, output_dim=4, num_seqs=1, seq_len=3)
    dataset.initialize()
    dataset.init_seq_order(epoch=1)
    engine = Engine(config=config)
    data_loader = engine._create_data_loader(dataset)
    for _ in data_loader:
        return
    assert False, "Should have contained sequences"


def test_data_loader_oggzip():
    from test_Dataset import create_ogg_zip_txt_only_dataset_mult_seqs

    ds_num_seqs = 23
    ds_max_seq_len = 11
    max_seqs = 3
    config = Config({"max_seqs": max_seqs, "batch_size": max_seqs * ds_max_seq_len})
    with create_ogg_zip_txt_only_dataset_mult_seqs(num_seqs=ds_num_seqs, max_seq_len=ds_max_seq_len) as dataset:
        dataset.init_seq_order(epoch=1)
        engine = Engine(config=config)
        data_loader = engine._create_data_loader(dataset)
        num_batches = 0
        num_seqs = 0
        last_batch_num_seqs = None
        for batch in data_loader:
            assert isinstance(batch, dict)
            data: torch.Tensor = batch["classes"]
            assert isinstance(data, torch.Tensor)
            num_batches += 1
            num_seqs += data.shape[0]
            if last_batch_num_seqs is not None:
                assert last_batch_num_seqs == max_seqs
            last_batch_num_seqs = data.shape[0]
        assert 1 <= last_batch_num_seqs <= max_seqs
        assert num_batches == -(-num_seqs // max_seqs) and num_seqs == ds_num_seqs

    ds_num_seqs = 5
    ds_max_seq_len = 5
    max_seqs = 2
    config = Config({"max_seqs": max_seqs, "batch_size": max_seqs * ds_max_seq_len})
    batches = []
    with create_ogg_zip_txt_only_dataset_mult_seqs(num_seqs=ds_num_seqs, max_seq_len=ds_max_seq_len) as dataset:
        dataset.init_seq_order(epoch=1)
        engine = Engine(config=config)
        data_loader = engine._create_data_loader(dataset)
        for batch in data_loader:
            assert isinstance(batch, dict)
            data: torch.Tensor = batch["classes"]
            batches.append(data.numpy().tolist())
    print(batches)
    # The following depends on the random data generation in create_ogg_zip_txt_only_dataset_mult_seqs,
    # but we fixed the seed and the random number generator, so this should stay the same, unless we change the code.
    assert batches == [[[12, 8, 9, 11], [16, 0, 0, 0]], [[6, 25, 18, 20, 5], [28, 10, 28, 14, 0]], [[17, 23]]]


def test_load_optimizer_old_format():
    config = Config(dict(optimizer={"class": "adamw", "weight_decay": 1e-3}))
    model = torch.nn.Linear(7, 5)
    updater = Updater(config=config, network=model, device=torch.device("cpu"))
    updater.create_optimizer()

    with tempfile.TemporaryDirectory(prefix="returnn_test_load_optimizer_old_format") as tmp_dir:
        torch.save(updater.optimizer.state_dict(), tmp_dir + "/model.opt.old_format.pt")
        updater.load_optimizer(tmp_dir + "/model.opt.old_format.pt")

        updater.save_optimizer(tmp_dir + "/model.opt.new_format.pt")
        updater.load_optimizer(tmp_dir + "/model.opt.new_format.pt")


def test_optimizer_convert_aux_param():
    # See rf_module_to_pt_module aux_params_as_buffers option.
    # This causes a change in the optimizer state dict.
    # But we should be able to convert it back, in both directions.

    from returnn.torch.frontend.bridge import rf_module_to_pt_module

    config = Config(dict(optimizer={"class": "adamw", "weight_decay": 1e-3}))
    rf.select_backend_torch()

    class _Model(rf.Module):
        def __init__(self):
            super().__init__()
            self.batch_norm = rf.BatchNorm(in_dim=rf.Dim(3))
            self.linear = rf.Linear(in_dim=rf.Dim(2), out_dim=rf.Dim(3))

    rf_model = _Model()
    pt_model_buf = rf_module_to_pt_module(rf_model, aux_params_as_buffers=True)
    pt_model_param = rf_module_to_pt_module(rf_model, aux_params_as_buffers=False)
    pt_model_buf_param_names = set(name for name, _ in pt_model_buf.named_parameters())
    pt_model_param_param_names = set(name for name, _ in pt_model_param.named_parameters())
    print("buf params:", pt_model_buf_param_names)
    print("all params:", pt_model_param_param_names)
    assert len(pt_model_buf_param_names) < len(pt_model_param_param_names)
    assert pt_model_buf_param_names.issubset(pt_model_param_param_names)
    updater_buf = Updater(config=config, network=pt_model_buf, device=torch.device("cpu"))
    updater_buf.create_optimizer()
    updater_param = Updater(config=config, network=pt_model_param, device=torch.device("cpu"))
    updater_param.create_optimizer()

    with tempfile.TemporaryDirectory(prefix="returnn_test_optimizer_convert_aux_param") as tmp_dir:
        updater_buf.save_optimizer(tmp_dir + "/model_buf.opt.pt")
        updater_param.save_optimizer(tmp_dir + "/model_param.opt.pt")
        updater_buf.load_optimizer(tmp_dir + "/model_buf.opt.pt")
        updater_param.load_optimizer(tmp_dir + "/model_param.opt.pt")
        # Ok, now test whether we can convert them.
        updater_buf.load_optimizer(tmp_dir + "/model_param.opt.pt")
        updater_param.load_optimizer(tmp_dir + "/model_buf.opt.pt")


class _DemoException(Exception):
    pass


class _TestTorchSubModelRaisingException(torch.nn.Module):
    def __init__(self, in_features: int, out_features: int):
        super().__init__()
        self.lin = torch.nn.Linear(in_features, out_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: [B,T,D]
        :return: [B,T,D']
        """
        x = self.lin(x)
        if int("1") == 1:
            raise _DemoException("uh")
        return x


def test_torch_engine_train_exception():
    class _Model(torch.nn.Module):
        def __init__(self, **_kwargs):
            super(_Model, self).__init__()
            self.sub = _TestTorchSubModelRaisingException(9, 2)

        def forward(self, x: torch.Tensor) -> torch.Tensor:
            """
            :param x: [B,T,D]
            :return: [B,T,D']
            """
            x = self.sub(x)
            return torch.nn.functional.log_softmax(x, dim=-1)

        @classmethod
        def train_step(cls, *, model: _Model, extern_data: TensorDict, **_kwargs):
            """train step"""
            data: Tensor = extern_data["data"]
            logits = model(data.raw_tensor)
            logits_packed = torch.nn.utils.rnn.pack_padded_sequence(
                logits, data.dims[1].dyn_size_ext.raw_tensor, batch_first=True, enforce_sorted=False
            )
            targets = extern_data["classes"]
            targets_packed = torch.nn.utils.rnn.pack_padded_sequence(
                targets.raw_tensor, data.dims[1].dyn_size_ext.raw_tensor, batch_first=True, enforce_sorted=False
            )
            loss = torch.nn.CrossEntropyLoss(reduction="none")(logits_packed.data, targets_packed.data.long())
            rf.get_run_ctx().mark_as_loss(name="ce", loss=loss)
            frame_error = torch.argmax(logits_packed.data, dim=-1).not_equal(targets_packed.data)
            rf.get_run_ctx().mark_as_loss(name="fer", loss=frame_error, as_error=True)

    config = Config(
        dict(
            task="train",
            device="cpu",
            extern_data={"data": {"dim": 9}, "classes": {"dim": 2, "sparse": True}},
            get_model=_Model,
            train_step=_Model.train_step,
            batch_size=500,
            optimizer={"class": "adam"},
        )
    )
    dataset = init_dataset({"class": "Task12AXDataset", "num_seqs": 100, "name": "train"})
    dataset.init_seq_order(epoch=1)

    with global_config_ctx(config):
        engine = Engine(config=config)
        engine.init_train_from_config(train_data=dataset)
        try:
            engine.train()
        except _DemoException as exc:
            print("got demo exception:", exc)
            exc_lines = str(exc).splitlines()
            assert "Module call stack:" in exc_lines and "(_TestTorchSubModelRaisingException.forward) sub" in exc_lines
        else:
            raise Exception("did not get expected exception")


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute
